/////////////////////////////////////////////////////////////////////////
//	"devtest.cpp"
//------------------------------------------------------------------------
//	$Id: devtest.cpp 199 2006-09-21 18:17:26Z taro $
//////////////////////////////////////////////////////////////////////////

#ifdef	HAVE_CONFIG_H
#	include	"config.h"
#	undef	HAVE_CONFIG_H
#endif
#include	<stdlib.h>
#include	<stdio.h>
#include	<string.h>
#include	<ctype.h>
#include	<time.h>
#if defined(WIN32)
# include	<crtdbg.h>
#elif defined(LINUX)
# include	<assert.h>
#endif
#include	"devtest.h"
#include	"partition.h"
#include	"tform.h"
#include	"getargopt.h"
#include	"defs.h"

#if	defined(LINUX)
#	define	_ASSERT
#endif

#if	!defined(HAVE_STRCMPI) && !defined(WIN32)
#	define	strcmpi		strcasecmp
#endif

//////////////////////////////////////////////////////////////////////////
//	DevTest

DevTest::DevTest()
{
	this->opt.cmd            = CMD_SCAN;
#if	defined(WIN32)
	this->opt.scanMode       = SCAN_DRIVE;
#endif
	this->opt.byteUnit       = BYTEUNIT_103;
	this->opt.printPartition = FALSE;
	this->opt.printDetail    = FALSE;
	this->opt.printDevInfo   = FALSE;
	this->opt.mbrtype        = MBRTYPE_UNKNOWN;
	this->opt.dumpModeSense  = FALSE;
	this->opt.writeEnable    = FALSE;
	this->opt.benchCommand   = FALSE;
	this->opt.benchSeek      = FALSE;
	this->opt.benchRead      = FALSE;
	this->opt.benchWrite     = FALSE;
	this->opt.benchRandomRW  = FALSE;
	this->opt.benchRwRatio   = 90 * RWRATIO_SCALE / 100;	// 90%
	this->opt.benchCsvResult = NULL;
	this->opt.measureTimeSec = 1;
	this->opt.addr           = I64(0);
	this->opt.count          = I64(0);
	this->opt.readAddr       = I64(0);
	this->opt.writeAddr      = I64(0);
	this->opt.ignoreError    = FALSE;
	this->opt.zeroFill       = FALSE;
	this->opt.retry          = 3;
	this->opt.searchBlock    = FALSE;
	this->opt.scsiCmd        = FALSE;
	this->opt.verify         = FALSE;
	this->opt.defBlockSize   = IBM_DISK_BLOCKSIZE;
	this->opt.cdblength      = DevCtl::SCSI_CDB12_LEN;
	this->opt.mode_cdb       = MODE_CDB6;
#if	defined(DEVTEST_DUMP_CSTYLE)
	this->opt.dumpCStyle     = FALSE;
#endif
	this->opt.filename       = NULL;
	this->opt.cpuUsage       = CPUUSAGE_OFF;
#if	defined(DEVTEST_USE_DISCSPEED)
	this->opt.discSpeed      = 1385;
#endif
	this->opt.testPath       = NULL;
	this->opt.testFileSize   = I64(0);
	this->opt.async          = FALSE;
	this->opt.nobuffer       = FALSE;
	this->opt.smart          = FALSE;
	this->opt.noprompt       = FALSE;

	this->ben.mode =
	    BEN_READ
	  | BEN_SEQUENTIAL
	  | BEN_RANDOM
	  | BEN_TRACK_START
	  | BEN_TRACK_END
	  | BEN_BLOCK_MIN
	  | BEN_BLOCK_MAX
	  | BEN_BLOCK_USER;
	this->ben.maxBufferSize   = MAX_BUFFER_SIZE;
	this->ben.testTimeSec     = TEST_TIME_SEC;
	this->ben.randomRWTimeSec = RND_TEST_TIME_SEC;
	this->ben.trackPos        = TRACK_POS_SCALE / 2;
	this->ben.blockSize       = 16384;
	this->ben.seekSector      = 0;	// default
	this->ben.discSpeed       = 0;
	this->ben.csvResult       = NULL;
	this->ben.read_func       = NULL;
	this->ben.write_func      = NULL;

	this->srcDevParam = NULL;
	this->dstDevParam = NULL;

	this->async.maxQueDepth   = 0;
	this->async.queNo         = 0;
	this->async.iotbl         = NULL;

//	this->medium.setStartStop    = FALSE;
//	this->medium.setLoadUnload   = FALSE;
//	this->medium.setPreventAllow = FALSE;
}

DevTest::~DevTest()
{
	if ( this->ben.csvResult != NULL && this->ben.csvResult != stdout && this->ben.csvResult != stderr )
		fclose(this->ben.csvResult);
	if ( this->async.iotbl != NULL )
		this->AsyncTerm();
	this->medium.clear();
}

#if	defined(WIN32)
//////////////////////////////////////////////////////////////////////////
//	Scan Device

int		DevTest::ScanDevice(DevTest::DevParam & devParam)
{
	switch ( this->opt.scanMode )
	{
		case SCAN_DRIVE:
			this->ScanDeviceDrive(devParam);
			break;
		case SCAN_DISK:
			this->ScanDeviceDisk(devParam);
			break;
		case SCAN_SCSI:
			this->ScanDeviceScsi(devParam);
			break;
#if	defined(DEVTEST_USE_CDROM)
		case SCAN_CDROM:
			this->ScanDeviceCdrom(devParam);
			break;
#endif
	}
	return 0;
}

// drive letter
int		DevTest::ScanDeviceDrive(DevTest::DevParam & devParam)
{
	int		drv;
	char	devname[MAX_PATH], desc[16];

	for ( drv = 'A'; drv <= 'Z'; drv++ )
	{
		sprintf(desc, "%c:", drv);
		sprintf(devname, "\\\\.\\%c:", drv);
		devParam.SetDesc(desc);
		devParam.SetDevName(devname);
		this->ScanDeviceCheck(devParam);
	}

	return 0;
}

// disk (physical drive
int		DevTest::ScanDeviceDisk(DevTest::DevParam & devParam)
{
	int		n = 0;
	char	devname[MAX_PATH], desc[16];

	for (;;)
	{
		sprintf(desc, "disk%d", n);
		sprintf(devname, "\\\\.\\PhysicalDrive%d", n);
		devParam.SetDesc(desc);
		devParam.SetDevName(devname);
		if ( this->ScanDeviceCheck(devParam) < 0 )
			break;
		n++;
	}

	return 0;
}

// SCSI
int		DevTest::ScanDeviceScsi(DevTest::DevParam & devParam)
{
	int		n = 0;
	char	devname[MAX_PATH], desc[16];

	for (;;)
	{
		sprintf(desc, "scsi%d", n);
		sprintf(devname, "\\\\.\\scsi%d:", n);
		devParam.SetDesc(desc);
		devParam.SetDevName(devname);
		if ( this->ScanDeviceCheck(devParam) < 0 )
			break;
		n++;
	}

	return 0;
}


#if	defined(DEVTEST_USE_CDROM)
// CD-ROM
int		DevTest::ScanDeviceCdrom(DevTest::DevParam & devParam)
{
	int		n = 0;
	char	devname[MAX_PATH], desc[16];

	for (;;)
	{
		sprintf(desc, "cdrom%d", n);
		sprintf(devname, "\\\\.\\CdRom%d", n);
		devParam.SetDesc(desc);
		devParam.SetDevName(devname);
		if ( this->ScanDeviceCheck(devParam) < 0 )
			break;
		n++;
	}

	return 0;
}
#endif

int		DevTest::ScanDeviceCheck(DevTest::DevParam & devParam)
{
	DevCtl::ScsiDevData 	*scsiDev = NULL;
	SCSI_ADDRESS			scsiAddr;
	BOOL					scsi_target = FALSE;

	if ( devParam.devname == NULL )
		return -1;
	this->devCtl.CloseDevice();
	devParam.ClearInqData();
	devParam.ClearCapacityData();
	this->devCtl.SetMaxCdbSize(this->opt.cdblength);
	if ( this->devCtl.OpenDevice(devParam.devname, DevCtl::RW_READ | DevCtl::RW_WRITE | DevCtl::RW_SHARE_READ | DevCtl::RW_SHARE_WRITE) < 0 )
		return -1;	// error

	if ( this->devCtl.GetHandleType() != DevCtl::HT_SCSI )
	{
		if ( this->devCtl.ScsiGetAddress(scsiAddr) == 0 )
			scsi_target = TRUE;
	}
	if ( (scsiDev = this->devCtl.ScsiGetInquiryData()) != NULL )
	{
		int			i;

		for ( i = 0; i < (int)scsiDev->GetBusNum(); i++ )
		{
			DevCtl::ScsiBusData 	*busData;
			int						j;

			if ( (busData = scsiDev->GetBusData(i)) == NULL )
				continue;
			for ( j = 0; j < (int)busData->GetInqNum(); j++ )
			{
				DevCtl::ScsiInqData 	*inqData;

				if ( (inqData = busData->GetInqData(j)) == NULL )
					continue;
				if ( scsi_target == FALSE
				  || (scsi_target != FALSE && scsiAddr.TargetId == inqData->GetTargetId() && scsiAddr.Lun == inqData->GetLun()) )
				{
					if ( this->opt.printPartition || this->opt.printDevInfo )
						fprintf(stdout,"\n");
					if ( scsi_target == FALSE && (i != 0 || j != 0) )
						devParam.SetDesc(NULL);
					devParam.SetInqData(inqData);
					this->PrintDevice(devParam);
				}
			}
		}
		_ASSERT(scsiDev);
		delete scsiDev;
	} else
	{
		DevCtl::ScsiInqData	inqData;
		unsigned			pathId, targetId, lun;

		if ( this->devCtl.GetFileType() != FILE_TYPE_DISK )
			return -1;
		if ( scsi_target )
		{
			pathId   = scsiAddr.PathId;
			targetId = scsiAddr.TargetId;
			lun      = scsiAddr.Lun;
		} else
		{
			pathId = targetId = lun = 0;
		}
		if ( this->devCtl.ScsiCreateInqData(inqData, pathId, targetId, lun) == 0 )
		{	// SCSI OK
			inqData.unknownId = TRUE;
			devParam.SetInqData(&inqData);
		} else
		{
			if ( this->opt.scsiCmd == FALSE )
				return -1;	// ERROR
		}
		this->PrintDevice(devParam);
	}

	return 0;
}
#endif

int	DevTest::OpenTargetDevice(DevCtl & devCtl, DevTest::DevTargetParam & tg, DevTest::DevParam & devParam, BOOL async)
{
	int						pathId, targetId, lun;
	unsigned				rwmode;

#if	defined(WIN32)
	tg.scsiDev = NULL;
	tg.busData = NULL;
#endif
	tg.inqData    = NULL;
	tg.setInqData = FALSE;

	if ( devParam.devname == NULL )
		return -1;
	pathId   = devParam.pathId;
	targetId = devParam.targetId;
	lun      = devParam.lun;
	devCtl.CloseDevice();
#if	defined(LINUX)
	devCtl.SetDefaultBlockSize(this->opt.defBlockSize);
#endif
	devParam.ClearInqData();
	devParam.ClearCapacityData();
	rwmode = DevCtl::RW_READ | DevCtl::RW_WRITE | DevCtl::RW_NONBLOCK | DevCtl::RW_SHARE_READ | DevCtl::RW_SHARE_WRITE;
	if ( this->IsAsyncMode() )
		rwmode |= DevCtl::RW_ASYNC;
	else
		rwmode |= DevCtl::RW_SYNC ;
	if ( this->opt.nobuffer )
		rwmode |= DevCtl::RW_NOBUFFER;
	devCtl.SetMaxCdbSize(this->opt.cdblength);
#if	defined(DEVTEST_USE_MD4RNET) && defined(DEVCTL_USE_MD4RNET)
	if ( devParam.devParamType == DEVPARAMTYPE_NET )
	{
		if ( devCtl.NetSet(devParam.net.host, devParam.net.port) < 0 )
		{
			fprintf(stderr,"%s - NetDevice open error!\n", devParam.devname);
			return -1;
		}
	}
#endif
	if ( devCtl.OpenDevice(devParam.devname, rwmode) < 0 )
	{
		rwmode = DevCtl::RW_READ | DevCtl::RW_NONBLOCK | DevCtl::RW_SHARE_READ | DevCtl::RW_SHARE_WRITE;
		if ( this->IsAsyncMode() )
			rwmode |= DevCtl::RW_ASYNC;
		else
			rwmode |= DevCtl::RW_SYNC ;
		if ( this->opt.nobuffer )
			rwmode |= DevCtl::RW_NOBUFFER;
		if ( devCtl.OpenDevice(devParam.devname, rwmode) < 0 )
		{
			fprintf(stderr,"%s - Device open error!\n", devParam.devname);
			return -1;	// error
		}
	}

#if	defined(WIN32)
	if ( pathId < 0 || targetId < 0 || lun < 0 )
	{
		SCSI_ADDRESS			scsiAddr;

		if ( devCtl.ScsiGetAddress(scsiAddr) == 0 )
		{	// Get SCSI Address
			pathId   = scsiAddr.PathId;
			targetId = scsiAddr.TargetId;
			lun      = scsiAddr.Lun;
		} else
		{
			pathId = targetId = lun = 0;
			if ( devCtl.ScsiCreateInqData(tg.inqDataWork, pathId, targetId, lun) < 0 )
			{	// error or Drive not ready
				fprintf(stderr,"%s - Device error!\n", devParam.GetDesc() != NULL ? devParam.GetDesc() : "" );
				return -1;
			}
			tg.inqData = &tg.inqDataWork;
			tg.inqData->unknownId = TRUE;
			tg.setInqData = TRUE;
		}
	}
	if ( tg.setInqData == FALSE )
	{
		if ( (tg.scsiDev = devCtl.ScsiGetInquiryData()) != NULL )
		{
			if ( (tg.busData = tg.scsiDev->GetBusData(pathId)) == NULL )
				return -1;
			for ( int j = 0; j < (int)tg.busData->GetInqNum(); j++ )
			{
				if ( (tg.inqData = tg.busData->GetInqData(j)) == NULL )
					continue;
				if ( (int)tg.inqData->GetPathId()   == pathId
				  && (int)tg.inqData->GetTargetId() == targetId
				  && (int)tg.inqData->GetLun()      == lun )
				{
					tg.setInqData = TRUE;
					break;
				}
			}
		} else
		{
			if ( devCtl.ScsiCreateInqData(tg.inqDataWork, pathId, targetId, lun) == 0 )
			{
				tg.inqData = &tg.inqDataWork;
				tg.inqData->unknownId = TRUE;
				tg.setInqData = TRUE;
			} else
			{	// NO SCSI
				if ( this->opt.scsiCmd == FALSE )
					return -1;
			}
		}
	}
#endif
#if	defined(LINUX)
	{
		if ( devCtl.ScsiCreateInqData(tg.inqDataWork, pathId, targetId, lun) < 0 )
			return -1;
		tg.inqData = &tg.inqDataWork;
		tg.inqData->unknownId = FALSE;
		tg.setInqData = TRUE;
	}
#endif

	if ( tg.setInqData != FALSE && tg.inqData != NULL )
	{
		devParam.SetInqData(tg.inqData);
	} else
	{
		if ( this->opt.scsiCmd == FALSE )
			return -1;
	}

	return 0;
}

void	DevTest::CloseTargetDevice(DevCtl & devCtl, DevTest::DevTargetParam & tg, DevTest::DevParam & devParam )
{
#if	defined(WIN32)
	if ( tg.scsiDev != NULL )
	{
		delete tg.scsiDev;
		tg.scsiDev = NULL;
	}
	devParam.ClearInqData();
#endif
	return;
}

//////////////////////////////////////////////////////////////////////////
//	Show,Dump,Benchmark,Surfscan: In DevTest.opt.cmd

int		DevTest::TargetDevice(DevTest::DevParam & devParam)
{
	DevTargetParam			tg;

	switch ( this->opt.cmd )
	{
		case CMD_SHOW     :
		case CMD_DUMP     :
			break;
		case CMD_BENCHMARK:
			if ( this->opt.benchCsvResult != NULL && this->ben.csvResult == NULL )
			{
				if ( strcmp(this->opt.benchCsvResult, "stdout") == 0 )
				{
					this->ben.csvResult = stdout;
				} else if ( strcmp(this->opt.benchCsvResult, "stderr") == 0 )
				{
					this->ben.csvResult = stderr;
				} else
				{
					if ( (this->ben.csvResult = fopen(this->opt.benchCsvResult, "a")) == NULL )
					{
						fprintf(stderr, "File open error! (%s)\n", this->opt.benchCsvResult);
						return -1;
					}
				}
			}
			break;
		case CMD_SURFSCAN :
			break;
		case CMD_READBLOCK :
			if ( this->opt.benchCsvResult != NULL && this->ben.csvResult == NULL )
			{
				if ( strcmp(this->opt.benchCsvResult, "stdout") == 0 )
				{
					this->ben.csvResult = stdout;
				} else if ( strcmp(this->opt.benchCsvResult, "stderr") == 0 )
				{
					this->ben.csvResult = stderr;
				} else
				{
					if ( (this->ben.csvResult = fopen(this->opt.benchCsvResult, "a")) == NULL )
					{
						fprintf(stderr, "File open error! (%s)\n", this->opt.benchCsvResult);
						return -1;
					}
				}
			}
			break;
#if	defined(DEVTEST_USE_DISCSPEED)
		case CMD_DISCSPEED:
			break;
#endif
		case CMD_MEDIUM:
			break;
		default:
			return -1;
	}
	if ( this->OpenTargetDevice(this->devCtl, tg, devParam, FALSE) == 0 )
	{
		switch ( this->opt.cmd )
		{
			case CMD_SHOW    :
				this->PrintDevice(devParam);
				break;
			case CMD_DUMP    :
				this->PrintDump(devParam);
				break;
			case CMD_BENCHMARK:
			{
				fprintf(stdout, "***** BENCHMARK TEST *****\n");
				if ( this->opt.testPath != NULL )
				{
					INT64	testSize;

					if ( (testSize = this->GetTestFileSize()) < 0 )
					{
						fprintf(stderr, "Test filesize error!\n");
						return -1;
					}
					DevTest::fprintf(stdout, "  FILE MODE %s (%llk : %,lld bytes)\n"
					  , this->opt.testPath
					  , testSize
					  , testSize);
				}
#if	defined(WIN32)
				if ( this->IsAsyncMode() )
					DevTest::fprintf(stdout, "  ACCESS MODE: Asynchronous\n");
#endif
				if ( this->BenchSetReadWrite() < 0 )
					return -1;
				fprintf(stdout, "\n");

				if ( this->opt.cpuUsage != CPUUSAGE_OFF )
					this->InitCpuUsage();
				this->opt.printPartition = FALSE;
				this->opt.printDetail    = FALSE;
				this->PrintDevice(devParam);

				if ( this->opt.testPath != NULL && (this->opt.benchCommand || this->opt.benchSeek) )
				{
					// 񓯊[hȂ炢N[Y
					if ( devParam.IsFileMode() && devParam.IsAsync() )
						this->TestFileClose(this->devCtl, devParam);
					if ( devParam.IsFileMode() == FALSE )
					{
						// [hŃI[v
						DevTest::fprintf(stderr, "  Create TestFile (%llu)\n", (INT64)this->GetTestFileSize());
						if ( this->TestFileOpen(this->devCtl, devParam, this->opt.testPath, this->GetTestFileSize(), FALSE) < 0 )
						{
							fprintf(stderr, "  TestFile open error! (%s)\n", this->opt.testPath);
							return -1;
						}
					}
				}
				if ( this->opt.benchCommand )
					this->BenchCommand( devParam, 0);
				if ( this->opt.benchSeek )
					this->BenchSeek( devParam, 0);

#if	defined(WIN32)
				if ( this->IsAsyncMode() )
				{	// re-open target device
					this->CloseTargetDevice(this->devCtl, tg, devParam);
					if ( this->OpenTargetDevice(this->devCtl, tg, devParam, TRUE) < 0 )
						return -1;
					if ( devParam.CheckCapacityData() == FALSE || devParam.GetCapacityBlockLength() == 0 )
					{
						devParam.ReadCapacityData(this->devCtl);
					}
				}
#endif
				if ( this->opt.testPath != NULL &&
				  (this->opt.benchRead || this->opt.benchWrite || this->opt.benchRandomRW) )
				{
					// ^񓯊[hwƈقȂ΂N[Y
					if ( devParam.IsFileMode() && devParam.IsAsync() != this->opt.async )
						this->TestFileClose(this->devCtl, devParam);
					if ( devParam.IsFileMode() == FALSE )
					{
						DevTest::fprintf(stderr, "  Create TestFile (%llu)\n", (INT64)this->GetTestFileSize());
						if ( this->TestFileOpen(this->devCtl, devParam, this->opt.testPath, this->GetTestFileSize(), this->opt.async) < 0 )
						{
							fprintf(stderr, "  TestFile open error! (%s)\n", this->opt.testPath);
							return -1;
						}
					}
				}
				if ( this->opt.benchRead )
				{
					MSLEEP(1000);
					this->BenchReadWrite( devParam, (this->ben.mode & ~BEN_WRITE) | BEN_READ);
				}
				if ( this->opt.benchWrite && (this->opt.testPath != NULL || this->opt.writeEnable) )
				{
					MSLEEP(1000);
					this->BenchReadWrite( devParam, (this->ben.mode & ~BEN_READ) | BEN_WRITE);
				}
				if ( this->opt.benchRandomRW && (this->opt.testPath != NULL || this->opt.writeEnable) )
				{
					MSLEEP(1000);
					this->BenchRandomReadWrite( devParam, this->opt.benchRwRatio);
				}
				if ( devParam.IsFileMode() )
					this->TestFileClose(this->devCtl, devParam);
				if ( this->opt.cpuUsage != CPUUSAGE_OFF )
					this->TermCpuUsage();
#if	defined(WIN32)
				if ( this->IsAsyncMode() )
					this->AsyncTerm();
#endif
				break;
			}
			case CMD_SURFSCAN:
			{
				this->SurfaceScan(devParam);
				break;
			}
			case CMD_READBLOCK:
				this->ReadWriteBlock(devParam);
				break;

#if	defined(DEVTEST_USE_DISCSPEED)
			case CMD_DISCSPEED:
				this->SetDiscSpeed(devParam);
				break;
#endif


			case CMD_MEDIUM:
			{
				MediumParam::iterator	it;
				MediumCommand			command;

				for ( it = medium.begin(); it != medium.end(); it++ )
				{
					command = *it;
					switch ( command.type )
					{
					case MEDIUM_PREVENT_ALLOW:
						fprintf(stderr, "Send Prevent/Allow Medium Removal Command %u\n", command.param);
						this->devCtl.ScsiPreventAllowMediumRemoval(
						   devParam.inqData->GetPathId()
						 , devParam.inqData->GetTargetId()
						 , devParam.inqData->GetLun()
						 , command.param
						);
						break;
					case MEDIUM_START_STOP:
						fprintf(stderr, "Send Start/Stop Unit Command %u\n", command.param);
						this->devCtl.ScsiStartStopUnit(
						   devParam.inqData->GetPathId()
						 , devParam.inqData->GetTargetId()
						 , devParam.inqData->GetLun()
						 , command.param
						);
						break;
					case MEDIUM_LOAD_UNLOAD:
						fprintf(stderr, "Send Load/Unload Medium Command %u\n", command.param);
						this->devCtl.ScsiLoadUnloadMedium(
						   devParam.inqData->GetPathId()
						 , devParam.inqData->GetTargetId()
						 , devParam.inqData->GetLun()
						 , command.param
						);
						break;
					}
				}
				break;
			}
		} // end of switch
	}
	this->CloseTargetDevice(this->devCtl, tg, devParam);

	return 0;
}

//////////////////////////////////////////////////////////////////////////
//	ShowDevice

int		DevTest::PrintDevice(DevTest::DevParam & devParam)
{
	BYTE						*data;
	DevCtl::SCSIINQUIRY 		*pinq = NULL;
	unsigned					pathId, targetId, lun;
	INT64						blkNum;
	unsigned					blkLen;
	char						sizeBuf[80];

	if ( devParam.CheckInqData()  )
	{
		pathId   = devParam.inqData->GetPathId();
		targetId = devParam.inqData->GetTargetId();
		lun      = devParam.inqData->GetLun();
		data     = devParam.inqData->GetData();
		pinq     = devParam.inqData->GetInquiryData();	// data == pinq
	} else
	{
		if ( this->opt.scsiCmd == FALSE )
			return -1;
	}

	{
		DevTest::DiskInf			diskinf;

		if ( this->GetDiskInf(this->devCtl, devParam, diskinf) == 0 )
		{
			if ( devParam.setDiskinf == FALSE )
			{
				devParam.diskinf    = diskinf;
				devParam.setDiskinf = TRUE;
			}
			blkNum = diskinf.blockNumber;
			blkLen = diskinf.blockSize;
			this->FormatSize(sizeBuf,  blkNum * blkLen);
		} else
		{
			blkNum = 0;
			blkLen = 0;
			sizeBuf[0] = '\0';
		}
	}
	devParam.ReadCapacityData(this->devCtl);

	//  ----+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8
	// "###### #-#-# ############################ - ####################### ### #######"
	// "DESC   P:I:L vender product version         device-type             FIX size"
	if ( devParam.GetDesc() != NULL && strlen(devParam.GetDesc()) > 6 )
		fprintf(stdout, "Device: %s\n", devParam.GetDesc());
	if ( pinq != NULL )
	{
		char	msg[256];

		sprintf(msg, "%-6s %X-%X-%X %-28.28s - %-23s %3s %7s\n"
		  , devParam.GetDesc(6)
		  , pathId
		  , targetId
		  , lun
		  , pinq->venderId
		  , this->devCtl.ScsiGetDeviceType(pinq->deviceType)
		  , pinq->removable != 0 ? "RMB" : "FIX"
		  , sizeBuf );
		fprintf(stdout, "%s", msg);
		if ( this->opt.benchCsvResult != NULL && this->ben.csvResult != NULL )
			fprintf(this->ben.csvResult, "%s", msg);
	} else 
	{	// NON SCSI
#if defined(WIN32)
		DISK_GEOMETRY				geo;
		char						msg[256], *dev_name, dev_name_buf[80], *fix_rem, *media_type, media_type_buf[80];

		if ( devCtl.DiskGetDriveGeometry(geo)  < 0 
		  && devCtl.StorageGetMediaTypes(geo)  < 0 
		  && devCtl.CdromGetDriveGeometry(geo) < 0
		)
		{
			return -1;
		}

		dev_name = NULL;
		fix_rem = "RMB";
		switch ( geo.MediaType )
		{
		case Unknown       : dev_name = "Unknown device"; fix_rem = "???"; break;
		case F5_1Pt2_512   : dev_name = "5.25\", 1.2 MB, 512B/sec"; break;
		case F3_1Pt44_512  : dev_name = "3.5\", 1.44 MB, 512B/sec"; break;
		case F3_2Pt88_512  : dev_name = "3.5\", 2.88 MB, 512B/sec"; break;
		case F3_20Pt8_512  : dev_name = "3.5\", 20.8 MB, 512B/sec"; break;
		case F3_720_512    : dev_name = "3.5\", 720 KB, 512B/sec"; break;
		case F5_360_512    : dev_name = "5.25\", 360 KB, 512B/sec"; break;
		case F5_320_512    : dev_name = "5.25\", 320 KB, 512B/sec"; break;
		case F5_320_1024   : dev_name = "5.25\", 320 KB, 1024B/sec"; break;
		case F5_180_512    : dev_name = "5.25\", 180 KB, 512B/sec"; break;
		case F5_160_512    : dev_name = "5.25\", 160 KB, 512B/sec"; break;
		case RemovableMedia: fix_rem = "RMB"; media_type = "Remobale Media"; break;
		case FixedMedia    : fix_rem = "FIX"; media_type = "Fixed Media"   ; break;
		case F3_120M_512   : dev_name = "3.5\", 120 MB media"; break;
		case F3_640_512    : dev_name = "3.5\", 640 KB, 512B/sec"; break;
		case F5_640_512    : dev_name = "5.25\", 640 KB, 512B/se"; break;
		case F5_720_512    : dev_name = "5.25\", 720 KB, 512B/sec"; break;
		case F3_1Pt2_512   : dev_name = "3.5\", 1.2 MB, 512B/sec"; break;
		case F3_1Pt23_1024 : dev_name = "3.5\", 1.23 MB, 1024B/sec"; break;
		case F5_1Pt23_1024 : dev_name = "5.25\", 1.23 MB, 1024B/sec"; break;
		case F3_128Mb_512  : dev_name = "3.5\" MO, 128 MB, 512 B/sec"; break;
		case F3_230Mb_512  : dev_name = "3.5\" MO, 230 MB, 512 B/sec"; break;
	//	case F8_256_12     : dev_name = "8\", 256 KB, 128B/sec"; break;
		default            : fix_rem = "???"; break;
		}
		if ( dev_name == NULL )
		{
			sprintf(dev_name_buf, "CHS %"FI64"u:%u:%u %u"
			  , (INT64)geo.Cylinders.QuadPart
			  , geo.TracksPerCylinder
			  , geo.SectorsPerTrack
			  , geo.BytesPerSector );
			dev_name = dev_name_buf;
		}
		if ( media_type == NULL )
		{
			sprintf(media_type_buf, "media-type %u", geo.MediaType);
			media_type = media_type_buf;
		}

		sprintf(msg, "%-6s ?-?-? %-28.28s - %-23s %3s %7s\n"
		  , devParam.GetDesc(6)
		  , dev_name
		  , media_type
		  , fix_rem
		  , sizeBuf );
		fprintf(stdout, "%s", msg);
		if ( this->opt.benchCsvResult != NULL && this->ben.csvResult != NULL )
			fprintf(this->ben.csvResult, "%s", msg);
#endif
	}

	if ( (devParam.CheckCapacityData() || this->opt.scsiCmd == FALSE )
	  && this->opt.printPartition
#if		defined(WIN32)
	  && (  this->devCtl.GetHandleType() == DevCtl::HT_DISK
	     || this->devCtl.GetHandleType() == DevCtl::HT_DRIVE )
#elif	defined(LINUX)
	  && (   this->devCtl.GetHType() == DevCtl::HT_BLKDEV
		  || this->devCtl.GetHType() == DevCtl::HT_CDROM
	      || pinq->deviceType == 0x00	// Direct Access Device
	      || pinq->deviceType == 0x05	// CD/DVD Device
	     )
#endif
	)
	{	// Partition
		this->PrintPartition(devParam);
	}
	if ( this->opt.printDevInfo )
	{	// Device
		this->PrintDevInfo(devParam);
	}

	if ( this->opt.smart )
	{	// S.M.A.R.T.
		this->PrintSmartInfo(devParam);
	}

	return 0;
}

//////////////////////////////////////////////////////////////////////////
//	Print Partition Table

int		DevTest::PrintPartition(DevTest::DevParam & devParam)
{
	BYTE						*buffer = NULL;
	INT64						blkNum;
	unsigned					blkLen;
	unsigned					blockSize;
	BOOL						cdrom = FALSE;

	{
		DevTest::DiskInf			diskinf;

		if ( this->GetDiskInf(this->devCtl, devParam, diskinf) < 0 )
			return -1;
		if ( devParam.setDiskinf == FALSE )
		{
			devParam.diskinf    = diskinf;
			devParam.setDiskinf = TRUE;
		}
		blkNum = diskinf.blockNumber;
		blkLen = diskinf.blockSize;
	}
	if ( devParam.CheckInqData() )
	{
		if ( devParam.CheckCapacityData() == FALSE || devParam.GetCapacityBlockLength() == 0 )
		{
			devParam.ReadCapacityData(this->devCtl);
		}
	}

	// CD-ROM
	if ( this->PrintCdrom(devParam) == 0 )
		cdrom = TRUE;

	if ( blkLen > DevTest::SECTOR_SIZE_MAX )
		return -1;
	blockSize = blkLen;
	if ( (buffer = new BYTE [blockSize]) == NULL )
		return -1;
	if ( this->ReadDev(this->devCtl, devParam, I64(0), 1, blockSize, buffer) < (int)blockSize )
		return -1;

	if ( blockSize >= IBM_DISK_BLOCKSIZE )
	{
		if ( this->opt.mbrtype == MBRTYPE_IBMFD
		  || (this->opt.mbrtype == MBRTYPE_UNKNOWN
		    && buffer[IBM_DISK_BLOCKSIZE-2] == 0x55 && buffer[IBM_DISK_BLOCKSIZE-1] == 0xAA
		    && (buffer[0] == 0xE9 || buffer[0] == 0xEB)) )
		{	// IBM floppy
			BPBDATA			*bpb;
			char			serial[80], fileSys[80];
			INT64			secVol;

			// "  IBM HARD-DISK FORMAT    (########) ########## sectors   ####################B
			DevTest::fprintf(stdout, "  IBM FLOPPY-DISK FORMAT  (%8u) %10"FI64"u sectors   %20,I64uB\n"
			  , blockSize
			  , (INT64)blkNum
			  , (INT64)blkNum * (INT64)blockSize
			);
			bpb = (BPBDATA *) &buffer[0x0B];
			if ( bpb->wRootEntry > 0 || bpb->wFatSector )
			{
				sprintf(serial, "%04X-%04X", bpb->dwSerialNo >> 16, bpb->dwSerialNo & 0xFFFF);
				sprintf(fileSys, "%.8s", bpb->sFileSystem);
				if ( (secVol = (INT64)bpb->wSectorVolume) == 0 )
					secVol = (INT64)bpb->dwLongSectorVolume;
				fprintf(stdout,"       %.11s %.9s %13"FI64"u .. %-24s %.8s\n"
				  , bpb->sVolumeLabel
				  , serial
				  , (INT64)secVol
				  , fileSys
				  , &buffer[0x03]
				);
				if ( this->opt.printDetail )
					this->PrintFatInfo(devParam, I64(0));
			} else
			{	// FAT32
				BF_BPB	*bfbpb;

				bfbpb = (BF_BPB *)bpb;
				sprintf(serial, "%04X-%04X", bfbpb->dwSerialNo >> 16, bfbpb->dwSerialNo & 0xFFFF);
				sprintf(fileSys, "%.8s", bfbpb->sFileSystem);
				if ( (secVol = (INT64)bfbpb->wSectorVolume) == 0 )
					secVol = (INT64)bpb->dwLongSectorVolume;
				fprintf(stdout,"       %.11s %.9s %13"FI64"u .. %-24s %.8s\n"
				  , bfbpb->sVolumeLabel
				  , serial
				  , (INT64)secVol
				  , fileSys
				  , &buffer[0x03]
				);
				if ( this->opt.printDetail )
					this->PrintFatInfo(devParam, I64(0));
			}
		} else if ( this->opt.mbrtype == MBRTYPE_IBMHD 
		  || (this->opt.mbrtype == MBRTYPE_UNKNOWN
		     && buffer[IBM_DISK_BLOCKSIZE-2] == 0x55 && buffer[IBM_DISK_BLOCKSIZE-1] == 0xAA
		     && blockSize == IBM_DISK_BLOCKSIZE) )
		{	// IBM HARDDISK
			int				ext;
			INT64			extAddr;

			//  ----+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8
			// "###### #-#-# ############################ - ####################### ### #######"
			// "### [### ## ##-### ## ##] ########## ########## ##:#################### #######
			// "  IBM HARD-DISK FORMAT    (########) ########## sectors   ####################B
			DevTest::fprintf(stdout, "  IBM HARD-DISK FORMAT    (%8u) %10"FI64"u sectors   %20,I64uB\n"
			  , blockSize
			  , (INT64)blkNum
			  , (INT64)blkNum * (INT64)blockSize
			);
			extAddr = 0;
			ext = 0;
			fprintf(stdout, "No  [Start CHS- End  CHS]   StartSec  SectorNum HH:Partition Type          SIZE\n");
			fprintf(stdout, "--- --------------------- ---------- ---------- --:-------------------- -------\n");
			this->PrintIbmPartition(devParam, ext, extAddr, I64(0), buffer);
		}
	} else if ( this->opt.mbrtype == MBRTYPE_MD4RAM
	  || (this->opt.mbrtype == MBRTYPE_UNKNOWN
	     && buffer[0] == 'M' && buffer[1] == 'D' && buffer[2] == '4' && buffer[3] == 'R' && buffer[4] == 'A' && buffer[5] == 'M') )
	{	// MD4RAM
		//  ----+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8
		// "  MD4RAM 0.99A#0 SIGN-BLK (########) ########## sectors   ####################B
		DevTest::fprintf(stdout, "  MD4RAM %-8.8s SIGN-BLK (%8u) %10"FI64"u sectors   %20,I64uB\n"
		  , &buffer[8]
		  , blockSize
		  , (INT64)blkNum
		  , (INT64)blkNum * (INT64)blockSize
		);
	} else if ( this->opt.mbrtype == MBRTYPE_FMRHD
	  || (this->opt.mbrtype == MBRTYPE_UNKNOWN && buffer[0] == 'I' && buffer[1] == 'P' && buffer[2] == 'L') )
	{	// FMR
		//  ----+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8
		// "  FMR HARD-DISK FORMAT    (########) ########## sectors   ####################B
		DevTest::fprintf(stdout, "  FMR HARD-DISK FORMAT    (%8u) %10"FI64"u sectors   %20,I64uB\n"
		  , blockSize
		  , (INT64)blkNum
		  , (INT64)blkNum * (INT64)blockSize
		);
	} else 
	{
#if	defined(DEVTEST_USE_UDFLIB)
		if ( this->PrintDisc(devParam) == 0 )
		{
		//	fprintf(stdout, "  UDF FORMAT DISC\n");
			cdrom = TRUE;
		}
#endif
		if ( cdrom == FALSE )
		{
		//  ----+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8
		// "  UNKNOWN FORMAT          (########) ########## sectors   ####################B
		DevTest::fprintf(stdout, "  UNKNOWN FORMAT          (%8u) %10"FI64"u sectors   %20,I64uB\n"
		  , blockSize
		  , (INT64)blkNum
		  , (INT64)blkNum * (INT64)blockSize
		);
		}
	}
	return 0;
}

int		DevTest::PrintIbmPartition(DevTest::DevParam & devParam, int & ext, INT64 & extAddr, INT64 baseAddr, BYTE * buffer)
{
	int				partNo;
	IBMPARTINF		partinf[4];
	INT64			blkNum;
	unsigned		blkLen;
	unsigned		blockSize;

	{
		DevTest::DiskInf			diskinf;

		if ( this->GetDiskInf(this->devCtl, devParam, diskinf) < 0 )
			return -1;
		if ( devParam.setDiskinf == FALSE )
		{
			devParam.diskinf    = diskinf;
			devParam.setDiskinf = TRUE;
		}
		blkNum = diskinf.blockNumber;
		blkLen = diskinf.blockSize;
	}

	blockSize = blkLen;

	memcpy( &partinf, buffer + 0x1BE, sizeof(partinf));
	for ( partNo = 0; partNo < 4; partNo++ )
	{
		BOOL			err;
		IBMPARTINF	   *ppart;
		char			sizeBuf[32], noBuf[8], partName[80];

		err = 0;
		ppart = &partinf[partNo];
		if ( ppart->dwRelativeSector == 0 )
			continue;
		this->FormatSize(sizeBuf,  (INT64)ppart->dwSectorNumber * blkLen);
		if ( ext == 0 )
			sprintf(noBuf, "%1d ", partNo + 1);
		else if ( ext < 9 )
			sprintf(noBuf, "e%1d", ext );
		else
			sprintf(noBuf, "%2d", ext );
		if ( (ppart->bPartType == 0x05 || ppart->bPartType == 0x0F) && ext > 0 )
			sprintf(partName, "Extended  %10"FI64"u", (INT64)baseAddr);
		else
			strcpy(partName, this->GetPartitionType(ppart->bPartType));
		if ( extAddr + ppart->dwRelativeSector >= blkNum
		  || extAddr + ppart->dwRelativeSector + ppart->dwSectorNumber > blkNum )
		{
			sprintf(partName, "Error     %10"FI64"u", (INT64)baseAddr);
			err = 1;
		}
		if ( (ppart->bPartType == 0x05 || ppart->bPartType == 0x0F) && ext && this->opt.printDetail == FALSE )
		{	// extended partition
			
		} else
		{
			//  ----+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8
			// "### [### ## ##-### ## ##] ########## ########## ##:#################### #######
			fprintf(stdout, "%1s%2s [%03X %02X %02X-%03X %02X %02X] %10u %10u %02X:%-20s %7s\n"
			  , ppart->bBootFlag ? "*" : "" 
			  , noBuf
			  , ppart->bStart[2] + ((ppart->bStart[1]&0xC0)<<2)
			  , ppart->bStart[0]
			  , ppart->bStart[1]&0x3F
			  , ppart->bEnd[2] + ((ppart->bEnd[1]&0xC0)<<2)
			  , ppart->bEnd[0]
			  , ppart->bEnd[1]&0x3F
			  , ppart->dwRelativeSector + (unsigned)baseAddr
			  , ppart->dwSectorNumber
			  , ppart->bPartType
			  , partName // err == 0 && ext == 0 ? this->GetPartitionType(ppart->bPartType) : errPart
			  , sizeBuf
			);
		}
		switch ( ppart->bPartType )
		{
			case 0x05:	/* extend partition	*/
			case 0x0F:	/* extend partition	*/
			{
				INT64	addr;

				if ( extAddr + ppart->dwRelativeSector >= blkNum )
					break;	// error
				addr = ppart->dwRelativeSector + extAddr;
				if ( extAddr == 0 )
					extAddr = addr;
				if ( this->ReadDev(this->devCtl, devParam, addr, 1, blockSize, buffer) < (int)blockSize )
					break;	// error
				ext++;
				this->PrintIbmPartition(devParam, ext, extAddr, addr, buffer);
				ext--;
				break;
			}
			case 0x01:	// FAT12
			case 0x04:	// FAT16(<32M)
			case 0x06:	// FAT16(>=32M)
			case 0x0B:	// FAT32
			case 0x0C:	// FAT32 LBA
			case 0x0E:	// FAT16 LBA
			case 0x11:	// Hidden FAT12
			case 0x14:	// Hidden FAT16 (<32M)
			case 0x16:	// Hidden FAT16 (>=32M)
			case 0x1B:	// Hidden FAT32 Hidden
			case 0x1C:	// Hidden FAT32 LBA
			case 0x1E:	// Hidden FAT16 LBA
				if ( this->opt.printDetail )
					this->PrintFatInfo(devParam, ppart->dwRelativeSector + baseAddr);
				break;

			case 0x07:	// NTFS
				if ( this->opt.printDetail )
					this->PrintNtfsInfo(devParam, ppart->dwRelativeSector + baseAddr);
				break;

			case 0x83:	// Linux ext2
				if ( this->opt.printDetail )
					this->PrintExt2Info(devParam, ppart->dwRelativeSector + baseAddr, ppart->dwSectorNumber);
				break;
		}
	}
	return 0;
}

int		DevTest::PrintFatInfo(DevTest::DevParam & devParam, INT64 addr)
{
	INT64						blkNum;
	unsigned					blkLen;
	BYTE						*buffer = NULL;
	int							retcode;

	{
		DevTest::DiskInf			diskinf;

		if ( this->GetDiskInf(this->devCtl, devParam, diskinf) < 0 )
			return -1;
		if ( devParam.setDiskinf == FALSE )
		{
			devParam.diskinf    = diskinf;
			devParam.setDiskinf = TRUE;
		}
		blkNum = diskinf.blockNumber;
		blkLen = diskinf.blockSize;
	}
	if ( (buffer = new BYTE [blkLen]) == NULL )
		return -1;
	try
	{
		char			t1[256], t2[256];
		void *			p;
	
		if ( this->ReadDev(this->devCtl, devParam, addr, 1, blkLen, buffer) < (int)blkLen )
			throw -1;
		p = &buffer[0x0B];
		if ( ((BPBDATA *)p)->wRootEntry > 0 || ((BPBDATA *)p)->wFatSector )
		{	// FAT16
			BPBDATA			*bpb;

			bpb = (BPBDATA *)p;
			PutPos( 4, "Jump Code"         , COLUMN1+2, Fmt(t1,"%02X %02X %02X", buffer[0],buffer[1],buffer[2]), 0);
			PutPos( 4, "OEM ID"            , COLUMN1+2, Fmt(t1,"%8.8s", &buffer[3]), 0);
			PutPos( 4, "Sector Size"       , COLUMN1  , Fmt(t1,"%10u", bpb->wSectorSize), 0);
			PutPos( 4, "Sector per Cluster", COLUMN1  , Fmt(t1,"%10u (%u bytes)", bpb->bAllocUnit, (unsigned)bpb->bAllocUnit * bpb->wSectorSize), 0);
			PutPos( 4, "Reserve Sector"    , COLUMN1  , Fmt(t1,"%10u", bpb->wReserveSector), 0);
			PutPos( 4, "FAT Number"        , COLUMN1  , Fmt(t1,"%10u", bpb->bFatNumber), 0);
			PutPos( 4, "Root Entry"        , COLUMN1  , Fmt(t1,"%10u", bpb->wRootEntry), 0);
			PutPos( 4, "Sector Volume"     , COLUMN1  , Fmt(t1,"%10u (%,I64u bytes)", bpb->wSectorVolume, (INT64)bpb->wSectorSize * bpb->wSectorVolume), 0);
			PutPos( 4, "Media Descriptor"  , COLUMN1+8, Fmt(t1,"%02XH", bpb->bMediaDescriptor), 0);
			PutPos( 4, "FAT Sector"        , COLUMN1  , Fmt(t1,"%10u", bpb->wFatSector), 0);
			PutPos( 4, "Sector per Track"  , COLUMN1  , Fmt(t1,"%10u", bpb->wSectorPerTrack), 0);
			PutPos( 4, "Head Number"       , COLUMN1  , Fmt(t1,"%10u", bpb->wHeadNumber), 0);
			PutPos( 4, "Hidden Sector"     , COLUMN1  , Fmt(t1,"%10u", bpb->dwHiddenSector), 0);
			PutPos( 4, "Long Sector Volume", COLUMN1  , Fmt(t1,"%10u (%,I64u bytes)", bpb->dwLongSectorVolume, (INT64)bpb->wSectorSize * bpb->dwLongSectorVolume), 0);
			PutPos( 4, "Serial No"         , COLUMN1+1, Fmt(t1,"%04X-%04X", bpb->dwSerialNo>>16, bpb->dwSerialNo & 0xFFFF), 0);
			PutPos( 4, "Volume Label"      , COLUMN1-1, Fmt(t1,"%11.11s", bpb->sVolumeLabel), 0);
			PutPos( 4, "File System"       , COLUMN1+2, Fmt(t1,"%8.8s", bpb->sFileSystem), 0);
		} else
		{	// FAT32
			BF_BPB			*bpb;

			bpb = (BF_BPB *)p;
			PutPos( 4, "Jump Code"         , COLUMN1+2, Fmt(t1,"%02X %02X %02X", buffer[0],buffer[1],buffer[2]), 0);
			PutPos( 4, "OEM ID"            , COLUMN1+2, Fmt(t1,"%8.8s", &buffer[3]), 0);
			PutPos( 4, "Sector Size"       , COLUMN1  , Fmt(t1,"%10u", bpb->wSectorSize), 0);
			PutPos( 4, "Sector per Cluster", COLUMN1  , Fmt(t1,"%10u (%u bytes)", bpb->bAllocUnit, bpb->bAllocUnit, (unsigned)bpb->bAllocUnit * bpb->wSectorSize), 0);
			PutPos( 4, "Reserve Sector"    , COLUMN1  , Fmt(t1,"%10u", bpb->wReserveSector), 0);
			PutPos( 4, "FAT Number"        , COLUMN1  , Fmt(t1,"%10u", bpb->bFatNumber), 0);
		//	PutPos( 4, "Root Entry"        , COLUMN1  , Fmt(t1,"%10u", bpb->wRootEntry), 0);
			PutPos( 4, "Sector Volume"     , COLUMN1  , Fmt(t1,"%10u (%,I64u bytes)", bpb->wSectorVolume, (INT64)bpb->wSectorSize * bpb->wSectorVolume), 0);
			PutPos( 4, "Media Descriptor"  , COLUMN1+8, Fmt(t1,"%02XH", bpb->bMediaDescriptor), 0);
		//	PutPos( 4, "FAT Sector"        , COLUMN1  , Fmt(t1,"%10u", bpb->wFatSector), 0);
			PutPos( 4, "Sector per Track"  , COLUMN1  , Fmt(t1,"%10u", bpb->wSectorPerTrack), 0);
			PutPos( 4, "Head Number"       , COLUMN1  , Fmt(t1,"%10u", bpb->wHeadNumber), 0);
			PutPos( 4, "Hidden Sector"     , COLUMN1  , Fmt(t1,"%10u", bpb->dwHiddenSector), 0);
			PutPos( 4, "Long Sector Volume", COLUMN1  , Fmt(t1,"%10u (%,I64u bytes)", bpb->dwLongSectorVolume, (INT64)bpb->wSectorSize * bpb->dwLongSectorVolume), 0);
			PutPos( 4, "Big FAT Sector"    , COLUMN1  , Fmt(t1,"%10u", bpb->dwBigFatSector), 0);
			PutPos( 4, "Ext Flags"         , COLUMN1+6, Fmt(t1,"%04XH", bpb->wExtFlags), 0);
			PutPos( 4, "FS Version"        , COLUMN1+6, Fmt(t1,"%04XH", bpb->wFsVer), 0);
			PutPos( 4, "Root Cluster"      , COLUMN1  , Fmt(t1,"%10u", bpb->dwRootCluster), 0);
			PutPos( 4, "FS Information"    , COLUMN1  , Fmt(t1,"%10u", bpb->wFsInfo), 0);
			PutPos( 4, "Backup Boot Sector", COLUMN1  , Fmt(t1,"%10u", bpb->wBackupBootSector), 0);
			PutPos( 4, Fmt(t2,"Drive No %s", (bpb->bDrvNo & 0x80) ? "(primary)":""), COLUMN1  , Fmt(t1,"%10u", bpb->bDrvNo & 0x7F), 0);
			PutPos( 4, "Boot Signature"    , COLUMN1+8, Fmt(t1,"%02XH", bpb->bBootSignature), 0);
			PutPos( 4, "Serial No"         , COLUMN1+1, Fmt(t1,"%04X-%04X", bpb->dwSerialNo>>16, bpb->dwSerialNo & 0xFFFF), 0);
			PutPos( 4, "Volume Label"      , COLUMN1-1, Fmt(t1,"%11.11s", bpb->sVolumeLabel), 0);
			PutPos( 4, "File System"       , COLUMN1+2, Fmt(t1,"%8.8s", bpb->sFileSystem), 0);
		}
		
		retcode = 0;
	}
	catch (int ret) { retcode = ret; }
	delete [] buffer;
	return retcode;
}

int		DevTest::PrintNtfsInfo(DevTest::DevParam & devParam, INT64 addr)
{
	INT64						blkNum;
	unsigned					blkLen;
	BYTE						*buffer = NULL;
	int							retcode;

	{
		DevTest::DiskInf			diskinf;

		if ( this->GetDiskInf(this->devCtl, devParam, diskinf) < 0 )
			return -1;
		if ( devParam.setDiskinf == FALSE )
		{
			devParam.diskinf    = diskinf;
			devParam.setDiskinf = TRUE;
		}
		blkNum = diskinf.blockNumber;
		blkLen = diskinf.blockSize;
	}
	if ( (buffer = new BYTE [blkLen]) == NULL )
		return -1;
	try
	{
		char			t1[256];
		NTFSBootBlock	*ntbb;

		if ( this->ReadDev(this->devCtl, devParam, addr, 1, blkLen, buffer) < (int)blkLen )
			throw -1;
		ntbb = (NTFSBootBlock *)buffer;
		PutPos( 4, "Jump Code"         , COLUMN1+2, Fmt(t1,"%02X %02X %02X", ntbb->jump[0],ntbb->jump[1],ntbb->jump[2]), 0);
		PutPos( 4, "OEM ID"            , COLUMN1+2, Fmt(t1,"%8.8s", ntbb->oem_id), 0);
		PutPos( 4, "Sector Size"       , COLUMN1  , Fmt(t1,"%10u", ntbb->wSectorSize), 0);
		PutPos( 4, "Sector per Cluster", COLUMN1  , Fmt(t1,"%10u (%u bytes)", ntbb->bAllocUnit, (unsigned)ntbb->bAllocUnit * ntbb->wSectorSize), 0);
		PutPos( 4, "Reserve Sector"    , COLUMN1  , Fmt(t1,"%10u", ntbb->wReserveSector), 0);
	//	PutPos( 4, "FAT Number"        , COLUMN1  , Fmt(t1,"%10u", ntbb->bFatNumber), 0);		// not use
	//	PutPos( 4, "Root Entry"        , COLUMN1  , Fmt(t1,"%10u", ntbb->wRootEntry), 0);		// not use
	//	PutPos( 4, "Sector Volume"     , COLUMN1  , Fmt(t1,"%10u", ntbb->wSectorVolume), 0);	// not use
		PutPos( 4, "Media Descriptor"  , COLUMN1+8, Fmt(t1,"%02XH", ntbb->bMediaDescriptor), 0);
	//	PutPos( 4, "FAT Sector"        , COLUMN1  , Fmt(t1,"%10u", ntbb->wFatSector), 0);		// not use
		PutPos( 4, "Sector per Track"  , COLUMN1  , Fmt(t1,"%10u", ntbb->wSectorPerTrack), 0);
		PutPos( 4, "Head Number"       , COLUMN1  , Fmt(t1,"%10u", ntbb->wHeadNumber), 0);
		PutPos( 4, "Hidden Sector"     , COLUMN1  , Fmt(t1,"%10u", ntbb->dwHiddenSector), 0);
	//	PutPos( 4, "Long Sector Volume", COLUMN1  , Fmt(t1,"%10u", ntbb->dwLongSectorVolume), 0);	// not use
	//	PutPos( 4, "Big FAT Sector"    , COLUMN1  , Fmt(t1,"%10u", ntbb->dwBigFatSector), 0);		// not use
		PutPos( 4, "Total Sectors"     , COLUMN1  , Fmt(t1,"%10"FI64"u (%,I64u bytes)", (INT64)ntbb->totalSectors, (INT64)ntbb->wSectorSize * ntbb->totalSectors), 0);
		PutPos( 4, "MFT Logical Cluster"       , COLUMN1  , Fmt(t1,"%10"FI64"u", ntbb->lcnMFT), 0);
		PutPos( 4, "Mirror MFT Logical Cluster", COLUMN1  , Fmt(t1,"%10"FI64"u", ntbb->lcnMFTMirror), 0);
		PutPos( 4, "File Record Segment", COLUMN1  , Fmt(t1,"%10u", ntbb->dwFileRecSeg), 0);
		PutPos( 4, "Index Block"        , COLUMN1  , Fmt(t1,"%10u", ntbb->dwIndexBlock), 0);
		PutPos( 4, "Volume Serial No"   , COLUMN1-7, Fmt(t1,"%08X-%08X", (unsigned)(ntbb->volumeSerialNo>>I64(32)), (unsigned)ntbb->volumeSerialNo & 0xFFFFFFFF), 0);
		PutPos( 4, "Checksum"           , COLUMN1+2, Fmt(t1,"%08XH", ntbb->dwChecksum), 0);
		retcode = 0;
	}
	catch (int ret) { retcode = ret; }
	delete [] buffer;
	return retcode;
}


int		DevTest::PrintCdrom(DevTest::DevParam & devParam)
{
	INT64						blkNum;
	unsigned					blkLen;
	DevCtl::SCSITOCDATA			toc;
	char						t1[80], t2[80];
	unsigned					n, tocLength;
	BOOL						msf = FALSE;

	{
		DevTest::DiskInf			diskinf;

		if ( this->GetDiskInf(this->devCtl, devParam, diskinf) < 0 )
			return -1;
		if ( devParam.setDiskinf == FALSE )
		{
			devParam.diskinf    = diskinf;
			devParam.setDiskinf = TRUE;
		}
		blkNum = diskinf.blockNumber;
		blkLen = diskinf.blockSize;
	}

	if ( devParam.CheckInqData() == FALSE || devParam.CheckCapacityData() == FALSE )
		return -1;
	if ( this->devCtl.ScsiReadToc(devParam.pathId, devParam.targetId, devParam.lun, msf, toc) < 0 )
		return -1;

	tocLength = SC16(toc.tocLength);
	//  ----+----1----+----2----+----3----+----4----+----5----+----6----+----7----+----8
	// "### [## ## ##]-[## ## ##] ########## ########## ##:#################### #######
	// "CD-ROM TOC DATA               Start:###       End:###
	// "   [##] ## Data track     ############# ############# Pre-emphasis  DigitCopy
	// "   [##] ## Audio (#ch          ##:##:##    ##########           ON         OK
	fprintf(stdout,"CD/DVD TOC DATA           %13s %13s\n"
	  , Fmt(t1,"Start:%u", toc.startTrack)
	  , Fmt(t2,"End:%u"  , toc.lastTrack )
	);
	fprintf(stdout, "   TrackNo TrackType          StartAddr     TrackSize Pre-emphasis DigitCopy\n");
	for ( n = 0; n < sizeof(toc.track)/sizeof(*toc.track); n++ )
	{
		char		trackNoBuf[40], *trackType, adrBuf[40], timeBuf[40];
		const char	*pe, *dcopy;
		INT64		adr, frame, nextadr;
		unsigned	ff, ss, mm;

		if ( n * sizeof(*toc.track) + 10 >= tocLength )
			break;
		if ( toc.track[n].no == 0xAA )
			strcpy(trackNoBuf, "--");
		else
			Fmt(trackNoBuf, "%2u", toc.track[n].no);
		if ( toc.track[n].adrctl & 0x04 )
		{	// DATA
			trackType = "Data track";
		} else
		{	// Audio
			if ( toc.track[n].adrctl & 0x08 )
				trackType = "Audio (4ch";
			else
				trackType = "Audio (2ch";
		}
		pe    = (toc.track[n].adrctl & 0x01) ? "ON" : "OFF";
		dcopy = (toc.track[n].adrctl & 0x02) ? "OK" : "";
		if ( msf == FALSE )
		{
			adr     = (INT64) SC32(toc.track[n  ].cdAddr);
			nextadr = (INT64) SC32(toc.track[n+1].cdAddr);
		} else
		{
			mm = toc.track[n].cdAddr[1];
			ss = toc.track[n].cdAddr[2];
			ff = toc.track[n].cdAddr[3];
			adr = (INT64)mm * I64(60)*I64(75) + (INT64)ss*I64(75) + (INT64)ff;
		//	sprintf(adrBuf,"%u:%u:%u:%u"
		//	  , toc.track[n].cdAddr[0]
		//	  , toc.track[n].cdAddr[1]
		//	  , toc.track[n].cdAddr[2]
		//	  , toc.track[n].cdAddr[3]
		//	);

			mm = toc.track[n+1].cdAddr[1];
			ss = toc.track[n+1].cdAddr[2];
			ff = toc.track[n+1].cdAddr[3];
			nextadr = (INT64)mm * I64(60)*I64(75) + (INT64)ss * I64(75) + (INT64)ff;
		//	sprintf(timeBuf,"%u:%u:%u:%u"
		//	  , toc.track[n+1].cdAddr[0]
		//	  , toc.track[n+1].cdAddr[1]
		//	  , toc.track[n+1].cdAddr[2]
		//	  , toc.track[n+1].cdAddr[3]
		//	);
		}
		frame = nextadr - adr;
		if ( toc.track[n].adrctl & 0x04 )
		{	// DATA Trsck
			sprintf(adrBuf , "%"FI64"u", (INT64)adr );
			sprintf(timeBuf, "%"FI64"u", (INT64)frame * blkLen);
		} else
		{	// AUDIO Track
			ff = (unsigned)(adr % I64(75));
			ss = (unsigned)((adr/I64(75)) % I64(60));
			mm = (unsigned)((adr/I64(75) / I64(60)));
			sprintf(adrBuf,"%2u:%02u:%02u", mm, ss, ff);

			ff = (unsigned)(frame % I64(75));
			ss = (unsigned)((frame/I64(75)) % I64(60));
			mm = (unsigned)((frame/I64(75) / I64(60)));
			sprintf(timeBuf,"%2u:%02u:%02u", mm, ss, ff);
		}
		// "CD-ROM TOC DATA            Start:###    End:###
		// "   [##] ## Data track     ########## ########## Pre-emphasis DigitCopy
		// "   [##] ## Audio (#ch       ##:##:## ##########           ON        OK
		fprintf(stdout,"   [%2s] %2u %-14s %13s %13s %12s %9s\n"
		  , trackNoBuf
		  , n
		  , trackType
		  , adrBuf
		  , timeBuf
		  , pe
		  , dcopy
		  );
	}
	return 0;
}


//////////////////////////////////////////////////////////////////////////
//	Dump

int		DevTest::PrintDump(DevParam & devParam)
{
	BYTE						*buffer;
	INT64						blkNum;
	unsigned					blkLen;

	fprintf(stdout,"***** Dump : LBA %"FI64"u *****\n\n", (INT64)this->opt.addr);
	{
		DevTest::DiskInf			diskinf;

		if ( this->GetDiskInf(this->devCtl, devParam, diskinf) < 0 )
			return -1;
		if ( devParam.setDiskinf == FALSE )
		{
			devParam.diskinf    = diskinf;
			devParam.setDiskinf = TRUE;
		}
		blkNum = diskinf.blockNumber;
		blkLen = diskinf.blockSize;
	}

	if ( (buffer = new BYTE [blkLen]) == NULL )
		return -1;
	if ( this->ReadDev(this->devCtl, devParam, (INT64)this->opt.addr, 1, blkLen, buffer) < (int)blkLen )
		return -1;
	this->PrintDumpHex((unsigned)(this->opt.addr * blkLen), blkLen, blkLen, buffer);

	delete [] buffer;
	return 0;
}


//////////////////////////////////////////////////////////////////////////
//	SurfaceScan

int		DevTest::SurfaceScan(DevParam & devParam)
{
	BYTE						*buffer = NULL;
	unsigned					bufferSize, bn, n;
	INT64						blkNum, addr, errnum = 0;
	unsigned					blkLen, count;
	time_t						tmStart, tmEnd, tmNow, tmLast;
	Clock_t						clkLast, clk;
	BOOL						cmdVerify = TRUE;
	const char					*usecmd = "Verify";

	fprintf(stdout,"***** Surface Scan *****\n\n");

	{
		DevTest::DiskInf			diskinf;

		if ( this->GetDiskInf(this->devCtl, devParam, diskinf) < 0 )
			return -1;
		if ( devParam.setDiskinf == FALSE )
		{
			devParam.diskinf    = diskinf;
			devParam.setDiskinf = TRUE;
		}
		blkNum = diskinf.blockNumber;
		blkLen = diskinf.blockSize;
	}
	bn         = (this->ben.maxBufferSize + blkLen - 1) / blkLen;
	bufferSize = bn * blkLen;
	if ( this->opt.verify  == FALSE
	  || this->opt.scsiCmd == FALSE
	  || devParam.CheckInqData() == FALSE
	  || this->devCtl.ScsiVerify(devParam.pathId, devParam.targetId, devParam.lun, I64(0), 1) < 0
	)
	{
		usecmd = "Read";
		cmdVerify = FALSE;
		if ( (buffer = new BYTE [bufferSize]) == NULL )
			return -1;
	}
	if ( this->opt.cpuUsage != CPUUSAGE_OFF )
	{
		this->InitCpuUsage();
		this->CollectCpuUsage();
	}
	time(&tmStart);
	tmLast = tmStart;
	clkLast = this->Clock();
	if ( (addr = this->opt.addr) >= blkNum )
		addr = blkNum - 1;
	count = 0;
	while ( addr < blkNum )
	{
		time(&tmNow);
		if ( difftime(tmNow, tmLast) > 1 )
		{
			char		speed[80], remain[80], cpuusage[80];
			unsigned	total_sec_per_blk, sec_per_blk, remSec;

			speed[0] = remain[0] = 0;
			clk = this->Clock();
			if ( clk - clkLast > 0 )
			{
				sec_per_blk = (unsigned)( (INT64)CLOCKS_PER_SEC * (INT64)count / (INT64)(clk - clkLast) );
				sprintf(speed, "%u.%1ukB/s"
				  , (unsigned) ( (INT64)sec_per_blk * (INT64)blkLen / I64(1000))
				  , (unsigned) (I64(10) * (INT64)sec_per_blk * (INT64)blkLen / I64(1000)) % 10
				);
				total_sec_per_blk = (unsigned)( (INT64)(addr - this->opt.addr) / (INT64)(difftime(tmNow, tmStart)) );
				if ( sec_per_blk > 0 && total_sec_per_blk > 0 )
				{
					sec_per_blk = (sec_per_blk + total_sec_per_blk) / 2;
					remSec = (unsigned)((blkNum - addr) / (INT64)sec_per_blk);
					this->FormatTimeRemain(remain, remSec);
				}
			}
			if ( this->opt.cpuUsage != CPUUSAGE_OFF )
			{
				this->CollectCpuUsage();
				sprintf(cpuusage, ":CPU%5.1f%%", (double)this->GetCpuUsage() );
			} else
				cpuusage[0] = '\0';
			fprintf(stderr, "%-6s%10"FI64"u/%"FI64"u(%3u.%1u%%)%14s%-9s(Remain%13s\r"
			  , usecmd
			  , (INT64)addr
			  , (INT64)blkNum
			  , (unsigned)(I64(100) * addr / blkNum)
			  , (unsigned)(I64(1000) * addr / blkNum) % 10
			  , speed
			  , cpuusage
			  , remain
			);
			tmLast = tmNow;
			clkLast = clk;
			count = 0;
		}
		{
			BOOL	err;
			int		retry;

			err = FALSE;
			retry = 0;
			n = ( addr + bn < blkNum ) ? bn : (unsigned)(blkNum - addr);
			if ( cmdVerify )
			{
				if ( this->devCtl.ScsiVerify(devParam.pathId, devParam.targetId, devParam.lun, addr, n) < 0 )
					err = TRUE;
			} else
			{
				if ( this->ReadDev(this->devCtl, devParam, addr, n, n * blkLen, buffer) < (int)(n * blkLen) )
					err = TRUE;
			}
			if ( err )
			{
				errnum++;
				retry++;
				fprintf(stdout, "Surface Scan Error! LBA %"FI64"u-%"FI64"u (count %u)\n"
				    , (INT64)addr
				    , (INT64)addr + n - 1
				    , retry
				);
				while ( retry <= this->opt.retry )
				{
					unsigned	m, err;

					err = 0;
					for ( m = 0; m < n; m++ )
					{
						err = 0;
						if ( cmdVerify )
						{
							if ( this->devCtl.ScsiVerify(devParam.pathId, devParam.targetId, devParam.lun, addr + m, 1) < 0 )
								err = TRUE;
						} else
						{
							if ( this->ReadDev(this->devCtl, devParam, addr + m, 1, blkLen, buffer) < (int)blkLen )
								err = TRUE;
						}
						if ( err )
						{
							err++;
							errnum++;
							fprintf(stdout, "Surface Scan Error! LBA %"FI64"u (count %u)\n"
							    , (INT64)addr + m
							    , retry
							);
						}
					}
					if ( err == 0 )
						break;	// OK
					retry++;
				}
			}
			if ( retry > this->opt.retry )
			{	// error
				fprintf(stdout, "Surface Scan Error! LBA %"FI64"u-%"FI64"u\n", (INT64)addr, (INT64)addr + n - 1);
				break;	// erro error exist
			}
		}
		addr += n;
		count += n;
	}
	time(&tmEnd);
	{
		unsigned	sec;
		char		tmbuf[32];

		sec = (unsigned) difftime(tmEnd, tmStart);
		fprintf(stdout, "\nSurface scan completed. %usec (%13s) : Error %"FI64"u \n"
		  , sec, this->FormatTimeRemain(tmbuf, sec)
		  , (INT64)errnum);
	}
	if ( this->opt.cpuUsage != CPUUSAGE_OFF )
		this->TermCpuUsage();
	if ( buffer != NULL )
		delete [] buffer;
	return 0;
}

//////////////////////////////////////////////////////////////////////////
//	Read/Write Block

int		DevTest::ReadWriteBlock(DevTest::DevParam & devParam)
{
	BYTE						*buffer = NULL;
	INT64						blkNum, startAddr, endAddr, maxCount;
	unsigned					blkLen;
	time_t						tmStart, tmEnd, tmNow, tmLast;
	Clock_t						clkLast, clk;
	int							retcode;
	FILE						*fp = NULL;
	const char					*usecmd = this->opt.cmd == DevTest::CMD_READBLOCK ? "Read" : "Write";

	fprintf(stdout,"***** %s Block *****\n\n", usecmd);

	{
		DevTest::DiskInf			diskinf;

		if ( this->GetDiskInf(this->devCtl, devParam, diskinf) < 0 )
			return -1;
		if ( devParam.setDiskinf == FALSE )
		{
			devParam.diskinf    = diskinf;
			devParam.setDiskinf = TRUE;
		}
		blkNum = diskinf.blockNumber;
		blkLen = diskinf.blockSize;
	}

	if ( (startAddr = this->opt.addr) >= blkNum )
		startAddr = blkNum - 1;
	maxCount = (this->opt.count == I64(0) || this->opt.count > blkNum) ? blkNum : this->opt.count;
	if ( startAddr + maxCount > blkNum )
		maxCount = blkNum - startAddr;
	endAddr = startAddr + maxCount - 1;

	try
	{
		unsigned	bufBlks, bufferSize;
		INT64		addr, lbacount;
		unsigned	count;

		bufBlks    = (this->ben.maxBufferSize + blkLen - 1) / blkLen;
		bufferSize = bufBlks * blkLen;
		if ( (buffer = new BYTE [bufferSize]) == NULL )
			return -1;

		fprintf(stdout,"  Start Address %10"FI64"u\n", (INT64)startAddr);
		fprintf(stdout,"  %-5s Count   %10"FI64"u\n", usecmd, (INT64)maxCount  );
		fprintf(stdout,"  Buffer size   %10u\n", (unsigned)bufferSize);
		if ( this->opt.noprompt == FALSE )
		{
			int		ret;

			fprintf(stderr, "Ready ? (Y/N) ");
			ret = DevTest::SelectConsole("YN");
			if ( ret != 'Y' )
				throw -1;
		}

		if ( this->opt.cmd == DevTest::CMD_READBLOCK )
		{	// disk to file(out)
			if ( this->opt.filename != NULL )
			{
				if ( (fp = fopen(this->opt.filename, "wb")) == NULL )
				{
					fprintf(stderr, "File open error! (%s)\n", this->opt.filename);
					throw -1;
				}
			} else
				fp = stdout;
		} else
		{	// file(in) to disk
			if ( this->opt.writeEnable == FALSE )
			{
				fprintf(stderr, "Disable WRITE function!\n");
				throw -1;
			}
			if ( this->opt.zeroFill )
			{
				fp = NULL;
				memset(buffer, 0, blkLen);
			} else
			{
				if ( this->opt.filename != NULL )
				{
					if ( (fp = fopen(this->opt.filename, "rb")) == NULL )
					{
						fprintf(stderr, "File open error! (%s)\n", this->opt.filename);
						throw -1;
					}
				} else
					fp = stdin;
			}
		}

		if ( this->opt.benchCsvResult != NULL && this->ben.csvResult != NULL )
		{
			fprintf( this->ben.csvResult, "time(sec), addr(LBA), addr(bytes), blocks/s, kB/s\n");
		}
		if ( this->opt.cpuUsage != CPUUSAGE_OFF )
		{
			this->InitCpuUsage();
			this->CollectCpuUsage();
		}
		time(&tmStart);
		tmLast = tmStart;
		clkLast = this->Clock();
		count = 0;
		addr = startAddr;
		lbacount = I64(0);
		for (;;)
		{
			time(&tmNow);
			if ( difftime(tmNow, tmLast) >= this->opt.measureTimeSec || lbacount == maxCount )
			{
				char		speed[80], remain[80], cpuusage[80];
				unsigned	total_sec_per_blk, sec_per_blk, remSec;

				speed[0] = remain[0] = 0;
				clk = this->Clock();
				if ( clk - clkLast > 0 )
				{
					double	dtime;

					sec_per_blk = (unsigned)( (INT64)CLOCKS_PER_SEC * (INT64)count / (INT64)(clk - clkLast) );
					sprintf(speed, "%u.%1u"
					  , (unsigned) ( (INT64)sec_per_blk * (INT64)blkLen / I64(1000))
					  , (unsigned) ((I64(10) * (INT64)sec_per_blk * (INT64)blkLen / I64(1000)) % I64(10))
					);
					if ( (dtime = difftime(tmNow, tmStart)) == 0.0 )
						dtime = 1.0;
					total_sec_per_blk = (unsigned)( (INT64)(lbacount) / (INT64)dtime );
					if ( sec_per_blk > 0 && total_sec_per_blk > 0 )
					{
						sec_per_blk = (sec_per_blk + total_sec_per_blk) / 2;
						remSec = (unsigned)((maxCount - lbacount) / (INT64)sec_per_blk);
						this->FormatTimeRemain(remain, remSec);
					}
				} else
				{
					strcpy(speed, "?.?");
					strcpy(remain, "????");
				}
				if ( this->opt.cpuUsage != CPUUSAGE_OFF )
				{
					this->CollectCpuUsage();
					sprintf(cpuusage, ":CPU%5.1f%%", (double)this->GetCpuUsage() );
				} else
					cpuusage[0] = '\0';
				fprintf(stderr, "%-6s%10"FI64"u/%"FI64"u(%3u.%1u%%)%10skB/s%-9s(Remain%13s\r"
				  , usecmd
				  , (INT64)addr
				  , (INT64)endAddr
				  , (unsigned)(I64(100)  * (lbacount) / maxCount)
				  , (unsigned)(I64(1000) * (lbacount) / maxCount) % 10
				  , speed
				  , cpuusage
				  , remain
				);
				if ( this->opt.benchCsvResult != NULL && this->ben.csvResult != NULL )
				{
					fprintf( this->ben.csvResult, "%u, %"FI64"u, %"FI64"u, %u, %s \n" 
					    , (unsigned)difftime(tmNow, tmStart)
					    , (INT64)addr
					    , (INT64)addr * blkLen
					    , sec_per_blk
					    , speed
					);
				}
				tmLast = tmNow;
				clkLast = clk;
				count = 0;
			}
			if ( lbacount >= maxCount )
				break;

			{
				unsigned	n;

				if ( lbacount + (INT64)bufBlks <= maxCount )
					n = bufBlks;
				else
					n = (unsigned)(maxCount - lbacount);
				if ( n == 0 )
					break;	// error
				if ( this->opt.cmd == DevTest::CMD_READBLOCK )
				{
					// Read
					if ( this->ReadDev(this->devCtl, devParam, addr, n, n * blkLen, buffer) < (int)(n * blkLen) )
					{
						if ( this->opt.ignoreError )
						{
							unsigned	blk_n;

							for ( blk_n = 0; blk_n < n; blk_n++ )
							{
								if ( this->ReadDev(this->devCtl, devParam, addr + blk_n, 1, blkLen, buffer + blkLen * blk_n) < (int)blkLen )
								{
									fprintf(stderr, "ScsiRead Error! (lba %"FI64"u)\n", (INT64)addr + blk_n);
								}
								if ( fwrite(buffer + blkLen * blk_n, 1, blkLen, fp) < blkLen )
								{
									fprintf(stderr, "File Write Error!\n");
									throw -1;
								}
							}
						} else
						{
							fprintf(stderr, "ScsiWrite Error! (lba %"FI64"u)\n", (INT64)addr);
							throw -1;
						}
					}
					if ( this->opt.searchBlock && n > 0 && blkLen >= IBM_DISK_BLOCKSIZE )
					{
						unsigned	blk_n;

						for ( blk_n = 0; blk_n < n; blk_n++ )
						{
							if ( buffer[blkLen * blk_n + IBM_DISK_BLOCKSIZE-2] == 0x55
							  && buffer[blkLen * blk_n + IBM_DISK_BLOCKSIZE-1] == 0xAA )
							{
								fprintf(stderr, "Found SuperBlock (lba %"FI64"u)\n", (INT64)addr + blk_n);
							}
						}
					}
					if ( fwrite(buffer, 1, n * blkLen, fp) < n * blkLen )
					{
						fprintf(stderr, "File Write Error!\n");
						throw -1;
					}
				} else
				{
					// Write
					if ( this->opt.zeroFill )
					{
					} else
					{
						if ( fread(buffer, 1, n * blkLen, fp) < n * blkLen )
						{
							fprintf(stderr, "File Read Error!\n");
							throw -1;
						}
					}
				}
				if ( n > 0 )
				{
					lbacount += n;
					addr += n;
					count += n;
				}
			}
		}
		time(&tmEnd);

		{
			unsigned	sec;
			char		tmbuf[32];

			sec = (unsigned) difftime(tmEnd, tmStart);
			fprintf(stdout, "\n%s completed. %usec (%13s) \n"
			  , usecmd
			  , sec, this->FormatTimeRemain(tmbuf, sec)
			);
		}
		retcode = 0;
	} catch (int code) { retcode = code; }

	if ( fp != NULL && fp != stdin && fp != stdout )
		fclose(fp);
	if ( buffer != NULL )
		delete [] buffer;
	if ( this->opt.cpuUsage != CPUUSAGE_OFF )
		this->TermCpuUsage();

	return retcode;
}

#if	defined(DEVTEST_USE_DISCSPEED)
//////////////////////////////////////////////////////////////////////////
//	SetDiscSpeed

int		DevTest::SetDiscSpeed(DevParam & devParam)
{
	BYTE									buffer[4096], *p, *end;
	unsigned								pathId, targetId, lun;
	UINT64									blkNum;
	unsigned								blkLen;
	unsigned								mlen = 0, bdlen = 0;	// block desc. length
	BOOL									cdb10 = this->opt.mode_cdb == MODE_CDB10 ? TRUE : FALSE;
	DevCtl::SCSI_MODE_PARAMETER_HEADER6		*mhead6;
	DevCtl::SCSI_MODE_PARAMETER_HEADER10	*mhead10;
	DevCtl::ScsiPagecode2A_Mms				*mms;
	BOOL									modeSense = FALSE;

	fprintf(stdout,"***** Set Disc Speed : %ukB/s *****\n\n", (unsigned)this->opt.discSpeed);

	if ( devParam.CheckInqData() == FALSE )
		return -1;
	pathId   = devParam.inqData->GetPathId();
	targetId = devParam.inqData->GetTargetId();
	lun      = devParam.inqData->GetLun();
	devParam.ReadCapacityData(this->devCtl, &blkLen, &blkNum);

	if ( this->devCtl.ScsiSetCdSpeed(pathId, targetId, lun, 0x00, this->opt.discSpeed) < 0 )
	{
		fprintf(stderr, "Scsi Set CD Speed Error!\n");
		return -1;
	}

	for (;;)
	{
		memset(buffer, 0, sizeof(buffer));
		if ( cdb10 )
		{
			if ( this->devCtl.ScsiModeSense(pathId, targetId, lun, FALSE, 0, 0x3F, sizeof(buffer), buffer) >= 0 )
			{
				mhead10 = (DevCtl::SCSI_MODE_PARAMETER_HEADER10 *) buffer;
				mlen  = SC16(mhead10->length    );
				bdlen = SC16(mhead10->blkDescLen);
			} else
			{
				fprintf(stderr, "Scsi ModeSense Error!\n");
				return -1;	// error
			}
		} else
		{
			if ( this->devCtl.ScsiModeSense(pathId, targetId, lun, FALSE, 0, 0x3F, 255, buffer) >= 0 )
			{
				mhead6 = (DevCtl::SCSI_MODE_PARAMETER_HEADER6 *) buffer;
				mlen  = mhead6->length;
				bdlen = mhead6->blkDescLen;
				p = buffer + sizeof(*mhead6) + bdlen;
				if ( mlen > 6 && p[0] == 0 && p[1] == 0 )
				{	// modeSense ucess, but data nothing
					cdb10 = TRUE;
					continue;
				}
			} else
			{
				cdb10 = TRUE;
				continue;	// retry
			}
		}
		break;
	}
	if ( cdb10 )
	{
		p = buffer + sizeof(*mhead10) + bdlen;
		end = buffer + sizeof(mhead10->length) + mlen;
	} else
	{
		p = buffer + sizeof(*mhead6) + bdlen;
		end = buffer + sizeof(mhead6->length) + mlen;
	}

	while ( p < end )
	{
		char	msg[80];

		if ( p[1] == 0 )
			break;
		if ( (p[0] & 0x3F) == 0x2A )
		{
			modeSense = TRUE;
			mms = (DevCtl::ScsiPagecode2A_Mms *) p;
			msg[0] = 0;
			PutPos(4, msg, 38," - MM Capabilities & Mechanical Status", 0);
			this->PrintPagecode2A(devParam, p[1] + 2, p );
			break;
		}
		p += (unsigned)p[1] + 2;
	}

	return 0;
}
#endif


//////////////////////////////////////////////////////////////////////////
//	CreatePartition

//////////////////////////////////////////////////////////////////////////
//	main()

enum	Option
{
	OPT_SCAN          = DevTest::CMD_SCAN      ,
	OPT_SHOW          = DevTest::CMD_SHOW      ,
	OPT_BENCHMARK     = DevTest::CMD_BENCHMARK ,
	OPT_DUMP          = DevTest::CMD_DUMP      ,
	OPT_SURFSCAN      = DevTest::CMD_SURFSCAN  ,
	OPT_READBLOCK     = DevTest::CMD_READBLOCK ,
	OPT_MEDIUM        = DevTest::CMD_MEDIUM     ,
	OPT_HELP          = DevTest::CMD_MAX       ,
	OPT_ASYNC         ,
	OPT_COMMAND_TEST  ,
	OPT_SEEK_TEST     ,
	OPT_DEV_INFO      ,
	OPT_PAT_INFO      ,
	OPT_VERBOSE       ,
	OPT_103           ,
	OPT_210           ,
	OPT_BLOCKSIZE     ,
	OPT_BLOCKTEST     ,
	OPT_CDBLENGTH     ,
	OPT_CDSPEED       ,
	OPT_CPUUSAGE      ,
	OPT_COUNT         ,
	OPT_CSVRESULT     ,
	OPT_DVDSPEED      ,
#if	defined(DEVTEST_DUMP_CSTYLE)
	OPT_DUMPCSTYLE    ,
#endif
	OPT_FILE          ,
	OPT_IGNORE_ERROR  ,
	OPT_MAXBUFSIZE    ,
	OPT_MAXQUEDEPTH   ,
	OPT_MBRTYPE       ,
	OPT_MEASTIME      ,
	OPT_MODE10        ,
	OPT_MODE6         ,
	OPT_MODESENSEDUMP ,
	OPT_NOBUFFER      ,
	OPT_NOPROMPT      ,
	OPT_RANDOM_RW     ,
	OPT_RANDOM_RW_TIME,
	OPT_SCSI          ,
	OPT_SEARCHBLOCK   ,
	OPT_SEEKSECTOR    ,
	OPT_SMART         ,
	OPT_SMART_ATTR    ,
	OPT_TESTFILE      ,
	OPT_TESTSIZE      ,
	OPT_TESTTIME      ,
	OPT_TRACKZONE     ,
	OPT_RETRY         ,
	OPT_VERIFY        ,
	OPT_WINFILE       ,
};

static void	help(GetOpt & getopt)
{
	int			i;

	static const char *msg1[] =
	{
		"usage: devtest [<device>] [<command>] [<option>]",
#if	defined(WIN32)
		"  <device>",
		"      <c>:               .. Drive Letter      (ex. c:)    : Scan ?:",
		"      disk[<n>]          .. Physical Drive    (ex. disk0) : without <n> Scan",
#if	defined(DEVTEST_USE_CDROM)
		"      cdrom[<n>]         .. CD/DVD Dtive      (ex. cdrom0): without <n> Scan",
#endif
		"      scsi[<n>]          .. SCSI Host Adapter (ex. scsi0) : without <n> Scan",
#endif
		"  <command>",
	};
	for ( i = 0; i < sizeof(msg1)/sizeof(*msg1); i++)
		puts(msg1[i]);

	static const GetOpt::HelpItem	opt1[] =
	{
		{ NULL, OPT_SCAN      , ""      , "Scan device"             },
		{ NULL, OPT_SHOW      , ""      , "Show deivce"             },
		{ NULL, OPT_BENCHMARK , "<type>", "Benchmark test\n  <type> all/command/seek/read/write/randomRW" },
		{ NULL, OPT_DUMP      , "<addr>", "Dump Sector"             },
		{ NULL, OPT_SURFSCAN  , "<addr>", "Surface Scan"            },
		{ NULL, OPT_READBLOCK , "<addr>", "Read  Block/Sector data" },
		{ NULL, OPT_MEDIUM    , "<m-cmd>", "Medium command\n  <m-cmd> abort,uneject,eject\n          spindown,spinup,load,unload\n          lock,unlock,prevent,allow"},
	};
	getopt.Help(sizeof(opt1)/sizeof(*opt1), opt1);

	static const char *msg2[] =
	{
		"  <option>",
	};
	for ( i = 0; i < sizeof(msg2)/sizeof(*msg2); i++)
		puts(msg2[i]);

	static const GetOpt::HelpItem	opt2[] =
	{
		{ NULL, OPT_ASYNC       , ""        , "Asynchronus mode  (Benchmark" },
		{ NULL, OPT_COMMAND_TEST, ""        , "Command test      (Benchmark" },
		{ NULL, OPT_SEEK_TEST   , ""        , "Seek test         (Benchmark" },
		{ NULL, OPT_DEV_INFO    , ""        , "Print Device Information (Scan, Show" },
		{ NULL, OPT_PAT_INFO    , ""        , "Print Partition          (Scan, Show" },
		{ NULL, OPT_VERBOSE     , ""        , "Print Detail             (Scan, Show" },
		{ NULL, OPT_103         , ""        , "k = 10^3 Bytes (default)"   },
		{ NULL, OPT_210         , ""        , "K = 2^10 Bytes"             },
		{ NULL, OPT_BLOCKSIZE   , "<n>"     , "Default Block(Sector) Size" },
		{ NULL, OPT_BLOCKTEST   , "<n>"     , "Set Blocksize     (Benchmark" },
		{ NULL, OPT_CDBLENGTH   , "<n>"     , "SCSI CDB Length" },
		{ NULL, OPT_CDSPEED     , ""        , "Report CD speed   (Benchmark" },
#if	defined(WIN32)
		{ NULL, OPT_CPUUSAGE    , "[<mode>]", "Report CPU %      (Benchmark,SurfScan\n  <mode> .. {self|all}" },
#elif	defined(LINUX)
		{ NULL, OPT_CPUUSAGE    , "[<mode>]", "Report CPU %      (Benchmark,SurfScan\n  <mode> .. {self|sys|user}" },
#endif
		{ NULL, OPT_COUNT       , "<n>"     , "Block R/W count, n>=1 or \"MAX\" (R/W Block" },
		{ NULL, OPT_DVDSPEED    , ""        , "Report DVD speed  (Benchmark" },
		{ NULL, OPT_FILE        , "<path>"  , "Read/Write file   (R/W Block" },
		{ NULL, OPT_MAXBUFSIZE  , "<n>"     , "Max Buffer Size   (Benchmark,Surfscan" },
		{ NULL, OPT_MAXQUEDEPTH , "<n>"     , "Max Que Depth     (Benchmark" },
		{ NULL, OPT_MBRTYPE     , "<type>"  , "MBR Type          (Sdcan, Show\n  <type> .. {auto|ibmfd|ibmhd|fmrhd|udf|md4ram}" },
		{ NULL, OPT_MEASTIME    , "<sec>"   , "Interval time     (R/W Block" },
		{ NULL, OPT_NOPROMPT    , ""        , "No prompt         (R/W Block" },
		{ NULL, OPT_SCSI        , ""        , "SCSI Command Mode" },
		{ NULL, OPT_SEARCHBLOCK , ""        , "Search SuperBlock (ReadBlock" },
		{ NULL, OPT_SEEKSECTOR  , "<n>"     , "Seek Sector/Track (Benchmark Seek Test" },
		{ NULL, OPT_SMART       , ""        , "S.M.A.R.T.               (Scan, Show"},
		{ NULL, OPT_SMART_ATTR  , "<id#>:<rawtype>:<text>" , "Add S.M.A.R.T. attribute (Scan, Show"},
		{ NULL, OPT_TESTFILE    , "<path>"  , "Test file         (Benchmark" },
		{ NULL, OPT_TESTSIZE    , "<n>"     , "Test filesize     (Benchmark" },
		{ NULL, OPT_TRACKZONE   , "<n%>"    , "Track Zone 1-99%  (Benchmark" },
		{ NULL, OPT_RETRY       , "<n>"     , "Error Retry Count (Surfscan" },
		{ NULL, OPT_VERIFY      , ""        , "Verify Comamnd    (Surfscan" },
	};
	getopt.Help(sizeof(opt2)/sizeof(*opt2), opt2);

	static const char *msg3[] =
	{
#if	defined(WIN32)
		"  Scan     : devtest [<device>] [<option>]",
#endif
		"  Show     : devtest  <device>  [<option>]",
		"  Benchmark: devtest  <device> --benchmark [<option>]",
#if	defined(WIN32)
		"    --benchmark --readBlock  .. Same --readBlock 0 --count MAX --file NUL",
#elif defined(LINUX)
		"    --benchmark --readBlock  .. Same --readBlock 0 --count MAX --file /dev/null",
#endif
		"  Dump     : devtest  <device> --dump <adr> [<option>]",
		"  Surfscan : devtest  <device> --surfscan [<adr>] [<option>]",
	};

	for ( i = 0; i < sizeof(msg3)/sizeof(*msg3); i++)
		puts(msg3[i]);
}

int		main(int argc, const char ** argv)
{
	DevTest							devTest;
	DevCtl							devctl;
	char							devtmp[512];
	GetOpt							getopt;
	int								cmd = DevTest::CMD_SCAN;
	DevTest::DevArgList				devArgList;
	DevTest::DevArgIterator			it;
	enum	ArgError
	{
		ARG_SUCCESS     ,
		ARG_ERROR_PARAM ,
		ARG_ERROR_GETOPT,
		ARG_ERROR       ,
	};
	ArgError		argError;
	char			errmsg[1024] = "";

	fprintf(stdout, DEVTEST_TITLE   "\n");
	fprintf(stdout, DEVTEST_VERSION "\n");
#if		defined(WIN32)
	devTest.opt.cmd = DevTest::CMD_SCAN;
#elif	defined(LINUX)
	devTest.opt.cmd = DevTest::CMD_SHOW;
#endif

	{
		static GetOpt::AddItemParam	tbl[] =
		{
			{ OPT_SCAN            , "scan"           , NULL, GetOpt::VAR_FLAG  , TRUE , { NULL,0,0} },
			{ OPT_SHOW            , "show"           , NULL, GetOpt::VAR_FLAG  , TRUE , { NULL,0,0} },
			{ OPT_BENCHMARK       , "benchmark"      , NULL, GetOpt::VAR_TEXT  , TRUE , {   "",0,0} },
			{ OPT_DUMP            , "dump"           , NULL, GetOpt::VAR_INT64 , FALSE, { NULL,0,0} },
			{ OPT_SURFSCAN        , "surfscan"       , NULL, GetOpt::VAR_INT64 , TRUE , { NULL,0,0} },
			{ OPT_READBLOCK       , "readBlock"      , NULL, GetOpt::VAR_INT64 , FALSE, { NULL,0,0} },

			{ OPT_MEDIUM          , "medium"         , NULL, GetOpt::VAR_TEXT  , TRUE , {   "",0,0} },
			{ OPT_HELP            , "help"           ,  "?", GetOpt::VAR_FLAG  , TRUE , { NULL,0,0} },
			{ OPT_ASYNC           , "async"          , NULL, GetOpt::VAR_FLAG  , TRUE , { NULL,0,0} },
			{ OPT_COMMAND_TEST    ,             NULL ,  "c", GetOpt::VAR_FLAG  , TRUE , { NULL,0,0} },
			{ OPT_SEEK_TEST       ,             NULL ,  "s", GetOpt::VAR_FLAG  , TRUE , { NULL,0,0} },
			{ OPT_DEV_INFO        ,             NULL ,  "d", GetOpt::VAR_FLAG  , TRUE , { NULL,0,0} },
			{ OPT_PAT_INFO        ,             NULL ,  "p", GetOpt::VAR_FLAG  , TRUE , { NULL,0,0} },
			{ OPT_VERBOSE         , "verbose"        ,  "v", GetOpt::VAR_FLAG  , TRUE , { NULL,0,0} },
			{ OPT_103             , "103"            , NULL, GetOpt::VAR_FLAG  , TRUE , { NULL,0,0} },
			{ OPT_210             , "210"            , NULL, GetOpt::VAR_FLAG  , TRUE , { NULL,0,0} },
			{ OPT_BLOCKSIZE       , "blockSize"      , NULL, GetOpt::VAR_INT   , FALSE, { NULL,0,0} },
			{ OPT_BLOCKTEST       , "blockTest"      , NULL, GetOpt::VAR_INT   , FALSE, { NULL,0,0} },
			{ OPT_CDBLENGTH       , "cdblength"      , NULL, GetOpt::VAR_INT   , FALSE, { NULL,0,0} },
			{ OPT_CDSPEED         , "cdspeed"        , NULL, GetOpt::VAR_FLAG  , TRUE , { NULL,0,0} },
			{ OPT_COUNT           , "count"          , NULL, GetOpt::VAR_TEXT  , FALSE, { NULL,0,0} },
			{ OPT_CPUUSAGE        , "cpuUsage"       , NULL, GetOpt::VAR_TEXT  , TRUE , { "all",0,0} },
			{ OPT_CSVRESULT       , "csvResult"      , NULL, GetOpt::VAR_TEXT  , TRUE , { NULL,0,0} },
			{ OPT_DVDSPEED        , "dvdspeed"       , NULL, GetOpt::VAR_FLAG  , TRUE , { NULL,0,0} },
#if	defined(DEVTEST_DUMP_CSTYLE)
			{ OPT_DUMPCSTYLE      , "dumpCStyle"      , NULL, GetOpt::VAR_FLAG , TRUE , { NULL,0,0} },
#endif
			{ OPT_FILE            , "file"           , NULL, GetOpt::VAR_TEXT  , FALSE, { NULL,0,0} },
			{ OPT_IGNORE_ERROR    , "ignoreError"    , NULL, GetOpt::VAR_FLAG  , TRUE , { NULL,0,0} },
			{ OPT_MAXBUFSIZE      , "maxBufferSize"  , NULL, GetOpt::VAR_INT   , FALSE, { NULL,0,0} },
			{ OPT_MAXQUEDEPTH     , "maxQueDepth"    , NULL, GetOpt::VAR_INT   , FALSE, { NULL,0,0} },
			{ OPT_MBRTYPE         , "mbrtype"        , NULL, GetOpt::VAR_TEXT  , FALSE, { NULL,0,0} },
			{ OPT_MEASTIME        , "measureTime"    , NULL, GetOpt::VAR_INT   , FALSE, { NULL,0,0} },
			{ OPT_MODE10          , "mode10"         , NULL, GetOpt::VAR_FLAG  , TRUE , { NULL,0,0} },
			{ OPT_MODE6           , "mode6"          , NULL, GetOpt::VAR_FLAG  , TRUE , { NULL,0,0} },
			{ OPT_MODESENSEDUMP   , "modeSenseDump"  , NULL, GetOpt::VAR_FLAG  , TRUE , { NULL,0,0} },
			{ OPT_NOBUFFER        , "nobuffer"       , NULL, GetOpt::VAR_FLAG  , TRUE , { NULL,0,0} },
			{ OPT_NOPROMPT        , "noprompt"       , NULL, GetOpt::VAR_FLAG  , TRUE , { NULL,0,0} },
			{ OPT_RANDOM_RW       , "randomReadWrite", NULL, GetOpt::VAR_FLAG  , TRUE , { NULL,0,0} },
			{ OPT_RANDOM_RW_TIME  , "randomRWTime"   , NULL, GetOpt::VAR_INT   , FALSE, { NULL,0,0} },
			{ OPT_SCSI            , "scsi"           , NULL, GetOpt::VAR_FLAG  , TRUE , { NULL,0,0} },
			{ OPT_SEARCHBLOCK     , "searchBlock"    , NULL, GetOpt::VAR_FLAG  , TRUE , { NULL,0,0} },
			{ OPT_SEEKSECTOR      , "seekSector"     , NULL, GetOpt::VAR_INT   , FALSE, { NULL,0,0} },
			{ OPT_SMART           , "smart"          , NULL, GetOpt::VAR_FLAG  , TRUE , { NULL,0,0} },
			{ OPT_SMART_ATTR      , "smart_attr"     , NULL, GetOpt::VAR_TEXT  , FALSE, { NULL,0,0} },
			{ OPT_TESTFILE        , "testfile"       , NULL, GetOpt::VAR_TEXT  , FALSE, { NULL,0,0} },
			{ OPT_TESTSIZE        , "testSize"       , NULL, GetOpt::VAR_INT64 , FALSE, { NULL,0,0} },
			{ OPT_TESTTIME        , "testTime"       , NULL, GetOpt::VAR_INT   , FALSE, { NULL,0,0} },
			{ OPT_TRACKZONE       , "trackZone"      , NULL, GetOpt::VAR_TEXT  , FALSE, { NULL,0,0} },
			{ OPT_RETRY           , "retry"          , NULL, GetOpt::VAR_INT   , FALSE, { NULL,0,0} },
			{ OPT_VERIFY          , "verify"         , NULL, GetOpt::VAR_FLAG  , TRUE , { NULL,0,0} },
			{ OPT_WINFILE         , "winfile"        , NULL, GetOpt::VAR_FLAG  , TRUE , { NULL,0,0} },
		};
		if ( getopt.AddItem(sizeof(tbl)/sizeof(*tbl), tbl) < 0 )
		{
			printf("Getopt AddItem Error!\n");
			return EXIT_FAILURE;
		}
	}

	try
	{
		int			an;

		an = 1;
		getopt.SetCurPos(an);
		while ( getopt.GetCurPos() < argc )
		{
			GetOpt::Item	*item;
			const char		*s;

			if ( (item = getopt.ParseArg(an, argc, argv)) == NULL )
			{
				if ( (s = getopt.last_param) != NULL )
				{
					DevTest::DevParam	*devParam;

					if ( (devParam = new DevTest::DevParam) == NULL )
						return EXIT_FAILURE;
					devParam->SetDesc(s);
					if ( (s = getopt.GetCurArg(argc, argv)) != NULL && isdigit(*s) )
					{	// <device> <scsi-id>[:<lun>]
						an++;
						devParam->SetTargetId( atoi(s) );
						devParam->SetLun(0);
						while ( *s )
						{	// host adapter #
							if ( isdigit(*s) == 0 )
								break;
							++s;
						}
						if ( *s == ':' || *s == '-' )
						{	// lun
							++s;
							if ( isdigit(*s) )
								devParam->SetLun ( atoi(s) );
						}
					}
					devArgList.push_back(devParam);
				} else if ( getopt.err_option != NULL || getopt.err_param != NULL )
				{	// error
					throw ARG_ERROR_GETOPT;
				} // else // End of Arg
			} else
			{
				switch ( item->usertype )
				{
				case OPT_SCAN:
					if ( item->var.flag )
						cmd = devTest.opt.cmd = DevTest::CMD_SCAN;
					break;

				case OPT_SHOW:
					if ( item->var.flag )
						cmd = devTest.opt.cmd = DevTest::CMD_SHOW;
					break;

				case OPT_BENCHMARK:
					cmd = devTest.opt.cmd = DevTest::CMD_BENCHMARK;
					s = item->var.text;
					if ( s != NULL )
					{
						if ( strcmpi(s, "all") == 0 )
						{
							devTest.opt.benchCommand  = TRUE;
							devTest.opt.benchSeek     = TRUE;
							devTest.opt.benchRead     = TRUE;
							devTest.opt.benchWrite    = TRUE;
							devTest.opt.benchRandomRW = TRUE;
						} else if ( strcmpi(s, "command") == 0 )
						{
							devTest.opt.benchCommand  = TRUE;
						} else if ( strcmpi(s, "seek") == 0 )
						{
							devTest.opt.benchSeek      = TRUE;
						} else if ( strcmpi(s, "read") == 0 )
						{
							devTest.opt.benchRead     = TRUE;
						} else if ( strcmpi(s, "write") == 0 )
						{
							devTest.opt.benchWrite    = TRUE;
						} else if ( strcmpi(s, "randomReadWrite") == 0 || strcmpi(s, "randomRW") == 0 )
						{
							devTest.opt.benchRandomRW = TRUE;
						} else
							devTest.opt.benchRead     = TRUE;
					} else
						devTest.opt.benchRead     = TRUE;
					break;

				case OPT_DUMP:
					cmd = devTest.opt.cmd = DevTest::CMD_DUMP;
					devTest.opt.addr = item->var.n64;
					break;


				case OPT_SURFSCAN:
					cmd = devTest.opt.cmd = DevTest::CMD_SURFSCAN;
					devTest.opt.addr = item->var.n64;
					break;


				case OPT_READBLOCK:
					devTest.opt.addr = item->var.n64;
					if (  devTest.opt.cmd == DevTest::CMD_BENCHMARK )
					{
						devTest.opt.count = 0;	// --count max
#if		defined(WIN32)
						devTest.opt.filename = "NUL";			// --file NUL
#elif	defined(LINUX)
						devTest.opt.filename = "/dev/null";	// --file /dev/null
#endif
					} else
					{
						if ( devTest.opt.count == 0 )
							devTest.opt.count = 1;
					}
					cmd = devTest.opt.cmd = DevTest::CMD_READBLOCK;
					break;


				case OPT_MEDIUM:
				{
					DevTest::MediumCommand	command;

					cmd = devTest.opt.cmd = DevTest::CMD_MEDIUM;
					s = item->var.text;
					if ( s != NULL )
					{
						if ( strcmpi(s, "abort") == 0 )
						{
							command.type  = DevTest::MEDIUM_LOAD_UNLOAD;
							command.param = DevTest::SCSI_LOADUNLOAD_ABORT;
						} else if ( strcmpi(s, "uneject") == 0 )
						{
							command.type  = DevTest::MEDIUM_LOAD_UNLOAD;
							command.param = DevTest::SCSI_LOADUNLOAD_LOAD;
						} else if ( strcmpi(s, "eject") == 0 )
						{
							command.type  = DevTest::MEDIUM_LOAD_UNLOAD;
							command.param = DevTest::SCSI_LOADUNLOAD_UNLOAD;

						} else if ( (strcmpi(s, "spindown") == 0 || strcmpi(s, "down") == 0) )
						{
							command.type  = DevTest::MEDIUM_START_STOP;
							command.param = DevTest::SCSI_STARTSTOPUNIT_SPINDOWN;
						} else if ( (strcmpi(s, "spinup") == 0 || strcmpi(s, "up") == 0) )
						{
							command.type  = DevTest::MEDIUM_START_STOP;
							command.param = DevTest::SCSI_STARTSTOPUNIT_SPINUP;
						} else if ( strcmpi(s, "load") == 0 )
						{
							command.type  = DevTest::MEDIUM_START_STOP;
							command.param = DevTest::SCSI_STARTSTOPUNIT_LOAD;
						} else if ( strcmpi(s, "unload") == 0 )
						{
							command.type  = DevTest::MEDIUM_START_STOP;
							command.param = DevTest::SCSI_STARTSTOPUNIT_UNLOAD;

						} else if ( strcmpi(s, "lock") == 0 )
						{
							command.type  = DevTest::MEDIUM_PREVENT_ALLOW;
							command.param = DevTest::SCSI_PREVENTALLOW_LOCK;
						} else if ( strcmpi(s, "unlock") == 0 )
						{
							command.type  = DevTest::MEDIUM_PREVENT_ALLOW;
							command.param = DevTest::SCSI_PREVENTALLOW_UNLOCK;
						} else if ( strcmpi(s, "prevent") == 0 )
						{
							command.type  = DevTest::MEDIUM_PREVENT_ALLOW;
							command.param = DevTest::SCSI_PREVENTALLOW_PERSISTENT_PREVENT;
						} else if ( strcmpi(s, "allow") == 0 )
						{
							command.type  = DevTest::MEDIUM_PREVENT_ALLOW;
							command.param = DevTest::SCSI_PREVENTALLOW_PERSISTENT_ALLOW;
						} else
						{
							sprintf(errmsg, "--medium error (%s)\n", s);
							throw ARG_ERROR;
						}
						devTest.medium.push_back(command);
					} else
					{
						command.type  = DevTest::MEDIUM_START_STOP;
						command.param = DevTest::SCSI_STARTSTOPUNIT_SPINUP;
						devTest.medium.push_back(command);
					}
					break;
				}

				case OPT_HELP:
					help(getopt);
					return EXIT_FAILURE;

				case OPT_ASYNC:
					devTest.opt.async = item->var.flag;
					if ( devTest.async.maxQueDepth == 0 )
						devTest.async.maxQueDepth = DevTest::IO_QUE_DEPTH;
					break;

				case OPT_COMMAND_TEST:
					devTest.opt.benchCommand = item->var.flag;
					break;

				case OPT_SEEK_TEST:
					devTest.opt.benchSeek = item->var.flag;
					break;

				case OPT_DEV_INFO:
					devTest.opt.printDevInfo = item->var.flag;
					break;

				case OPT_PAT_INFO:
					devTest.opt.printPartition = item->var.flag;
					break;

				case OPT_VERBOSE:
					devTest.opt.printDetail = item->var.flag;
					break;

				case OPT_103:
					if ( item->var.flag )
						devTest.opt.byteUnit = DevTest::BYTEUNIT_103;
					break;

				case OPT_210:
					devTest.opt.byteUnit = DevTest::BYTEUNIT_210;
					break;

				case OPT_BLOCKSIZE:
					devTest.opt.defBlockSize = (unsigned) item->var.n;
					break;

				case OPT_BLOCKTEST:
					devTest.ben.blockSize = item->var.n;
					if ( devTest.ben.blockSize < DevTest::MIN_BUFFER_SIZE || devTest.ben.blockSize >= devTest.ben.maxBufferSize )
					{
						sprintf(errmsg, "Blocksize error!\n");
						throw ARG_ERROR;
					}
					devTest.ben.mode |= DevTest::BEN_BLOCK_USER;
					break;

				case OPT_CDBLENGTH:
					devTest.opt.cdblength = (unsigned) item->var.n;
					switch ( devTest.opt.cdblength )
					{
					case DevCtl::SCSI_CDB6_LEN:
					case DevCtl::SCSI_CDB10_LEN:
					case DevCtl::SCSI_CDB12_LEN:
					case DevCtl::SCSI_CDB16_LEN:
						break;
					default:
						sprintf(errmsg, "CDB Length error!\n");
						throw ARG_ERROR;
					}
					break;

				case OPT_CDSPEED:
					if ( item->var.flag )
						devTest.ben.discSpeed = DevTest::CD_SPEED;
					break;

				case OPT_COUNT:
					s = item->var.text;
					if ( strcmpi(s, "max") == 0 )
					{
						devTest.opt.count = 0;
					} else 
					{
						devTest.opt.count = GetOpt::GetInt64(s);
						if ( devTest.opt.count <= 0 )
						{
							sprintf(errmsg, "Too small count!\n");
							throw ARG_ERROR;
						}
					}
					break;

				case OPT_CPUUSAGE:
					devTest.opt.cpuUsage = DevTest::CPUUSAGE_SELF;
					s = item->var.text;
					_ASSERT(s != NULL);
					if ( strcmpi(s, "self") == 0 )
					{
						devTest.opt.cpuUsage = DevTest::CPUUSAGE_SELF;
					}
#if	defined(WIN32)
					else if ( strcmpi(s, "all") == 0 )
					{
						devTest.opt.cpuUsage = DevTest::CPUUSAGE_ALL;
					}
#endif
#if	defined(LINUX)
					else if ( strcmpi(s, "sys") == 0 )
					{
						devTest.opt.cpuUsage = DevTest::CPUUSAGE_SYS;
					} else if ( strcmpi(s, "user") == 0 )
					{
						devTest.opt.cpuUsage = DevTest::CPUUSAGE_USER;
					}
#endif
					break;

				case OPT_CSVRESULT:
					s = item->var.text;
					if ( s != NULL )
						devTest.opt.benchCsvResult = s;
					else
						devTest.opt.benchCsvResult = "stdout";
					break;

				case OPT_DVDSPEED:
					if ( item->var.flag )
						devTest.ben.discSpeed = DevTest::DVD_SPEED;
					break;

#if	defined(DEVTEST_DUMP_CSTYLE)
				case OPT_DUMPCSTYLE:
					devTest.opt.dumpCStyle = item->var.flag;
					break;
#endif

				case OPT_FILE:
					s = item->var.text;
					_ASSERT(s != NULL);
					devTest.opt.filename = s;
					break;

				case OPT_IGNORE_ERROR:
					devTest.opt.ignoreError = item->var.flag;
					break;

				case OPT_MAXBUFSIZE:
					devTest.ben.maxBufferSize = (unsigned) item->var.n;
					if ( devTest.ben.maxBufferSize < DevTest::MIN_BUFFER_SIZE )
					{
						sprintf(errmsg, "Too small MaxBufferSize!\n");
						throw ARG_ERROR;
					}
					break;

#if	defined(WIN32)
				case OPT_MAXQUEDEPTH:
					devTest.async.maxQueDepth = item->var.n;
					if ( devTest.async.maxQueDepth < 1 )
					{
						sprintf(errmsg, "Too small maxQueDepth!\n");
						throw ARG_ERROR;
					}
					break;
#endif

				case OPT_MBRTYPE:
					s = item->var.text;
					_ASSERT(s != NULL);
					if ( strcmpi(s, "auto") == 0 )
						devTest.opt.mbrtype = DevTest::MBRTYPE_UNKNOWN;
					else if ( strcmpi(s, "ibmfd") == 0 )
						devTest.opt.mbrtype = DevTest::MBRTYPE_IBMFD;
					else if ( strcmpi(s, "ibmhd") == 0 )
						devTest.opt.mbrtype = DevTest::MBRTYPE_IBMHD;
					else if ( strcmpi(s, "fmrhd") == 0 )
						devTest.opt.mbrtype = DevTest::MBRTYPE_FMRHD;
					else if ( strcmpi(s, "udf") == 0 )
						devTest.opt.mbrtype = DevTest::MBRTYPE_UDF;
					else if ( strcmpi(s, "md4ram") == 0 )
						devTest.opt.mbrtype = DevTest::MBRTYPE_MD4RAM;
					else
					{
						sprintf(errmsg, "MBR type error!\n");
						throw ARG_ERROR;
					}
					break;

				case OPT_MEASTIME:
					devTest.opt.measureTimeSec = item->var.n;
					if ( devTest.opt.measureTimeSec < 1 )
					{
						sprintf(errmsg, "Too small measureTime!\n");
						throw ARG_ERROR;
					}
					break;

				case OPT_MODE10:
					if ( item->var.flag )
						devTest.opt.mode_cdb = DevTest::MODE_CDB10;
					break;

				case OPT_MODE6:
					if ( item->var.flag )
						devTest.opt.mode_cdb = DevTest::MODE_CDB6;
					break;

				case OPT_MODESENSEDUMP:
					devTest.opt.dumpModeSense = item->var.flag;
					break;

				case OPT_NOBUFFER:
					devTest.opt.nobuffer = item->var.flag;
					break;

				case OPT_NOPROMPT:
					devTest.opt.noprompt = item->var.flag;
					break;

				case OPT_RANDOM_RW:
					devTest.opt.benchRandomRW = item->var.flag;
					break;

				case OPT_RANDOM_RW_TIME:
					devTest.ben.randomRWTimeSec = item->var.n;
					if ( devTest.ben.randomRWTimeSec < 1 )
					{
						sprintf(errmsg, "Too small rndTestTime!\n");
						throw ARG_ERROR;
					}
					break;


				case OPT_RETRY:
					devTest.opt.retry = item->var.n;
					break;

				case OPT_SCSI:
					devTest.opt.scsiCmd = item->var.flag;
					break;

				case OPT_SEARCHBLOCK:
					devTest.opt.searchBlock = item->var.flag;
					break;

				case OPT_SEEKSECTOR:
					devTest.ben.seekSector = item->var.n;
					break;

				case OPT_SMART:
					devTest.opt.smart = item->var.flag;
					break;

				case OPT_SMART_ATTR:
				{
					static const char			*sep = ":";
					char						*buffer, *token;
					unsigned					id = 0;
					DevTest::SmartAttrRawType	type = DevTest::SMART_RAW_NORMAL;
					char						*desc = NULL;
					DevTest::SmartAttrItem		*attr_item;

					if ( (buffer = new char [strlen(item->var.text) + 1]) == NULL )
						throw ARG_ERROR;
					strcpy(buffer, item->var.text);
					if ( (token = strtok( buffer, sep)) != NULL )
						id = atoi(token);
					if ( id  == 0 || id > 255 )
					{
						sprintf(errmsg, "S.M.A.R.T. --smart_attr attribute ID error!\n");
						throw ARG_ERROR;
					}
					if ( (token = strtok( NULL, sep)) != NULL )
					{
						if ( strcmpi(token, "normal") == 0 )
						{
							type = DevTest::SMART_RAW_NORMAL;
						} else if ( strcmpi(token, "divword") == 0 )
						{
							type = DevTest::SMART_RAW_DIVWORD;
						} else if ( strcmpi(token, "dumphex") == 0 )
						{
							type = DevTest::SMART_RAW_DUMPHEX;
						} else
						{
							sprintf(errmsg, "S.M.A.R.T. --smart_attr type error!\n");
							throw ARG_ERROR;
						}
					}
					if ( (token = strtok( NULL, sep)) != NULL )
						desc = token;
					if ( (attr_item = new DevTest::SmartAttrItem) == NULL )
						throw ARG_ERROR;
					attr_item->id = id;
					attr_item->type = type;
					attr_item->desc= desc;
					devTest.smart.attrList.push_back(attr_item);
					break;
				}

				case OPT_TESTFILE:
					s = item->var.text;
					_ASSERT(s != NULL);
					devTest.opt.testPath = s;
					if ( devTest.opt.scsiCmd )
						devTest.opt.scsiCmd = FALSE;
					break;

				case OPT_TESTSIZE:
					devTest.opt.testFileSize = item->var.n;
					if ( devTest.opt.testFileSize < (INT64)DevTest::TESTFILE_MINSIZE )
					{
						sprintf(errmsg, "Too small testFileSize!\n");
						throw ARG_ERROR;
					}
					break;

				case OPT_TESTTIME:
					devTest.ben.testTimeSec = item->var.n;
					if ( devTest.ben.testTimeSec < 1 )
					{
						sprintf(errmsg, "Too small testTime!\n");
						throw ARG_ERROR;
					}
					break;

				case OPT_TRACKZONE:
					s = item->var.text;
					devTest.ben.trackPos = DevTest::TRACK_POS_SCALE * atoi(s) / 100;
					if ( devTest.ben.trackPos < 0 || devTest.ben.trackPos >= DevTest::TRACK_POS_SCALE )
					{
						sprintf(errmsg, "Track Zone error!\n");
						throw ARG_ERROR;
					}
					devTest.ben.mode |= DevTest::BEN_TRACK_USER;
					break;

				case OPT_VERIFY:
					devTest.opt.verify = item->var.flag;
					break;


				default:
					_ASSERT(0);
					break;
				} // end of switch
			}
		} // end of while
		argError = ARG_SUCCESS;
	} catch(ArgError code) { argError = code; }
	if ( argError != ARG_SUCCESS )
	{
		switch ( argError )
		{
		case ARG_ERROR_PARAM:
			fprintf(stderr, "Parametr error! (%s)\n", getopt.GetLastPos() >= 0  ? argv[getopt.GetLastPos()] : "" );
			break;
		case ARG_ERROR_GETOPT:
			fprintf(stderr, "Option parameter error!\n");
			if ( getopt.err_option != NULL )
				fprintf(stderr, "  option   =\"%s\"\n", getopt.err_option);
			if ( getopt.err_param != NULL )
				fprintf(stderr, "  parameter=\"%s\"\n", getopt.err_param);
			break;
		case ARG_ERROR:
		default:
			fputs(errmsg, stderr);
			break;
		}
		return EXIT_FAILURE;
	}

#if	defined(WIN32)
	if ( devArgList.size() == 0 )
	{
		DevTest::DevParam	*devParam;

		if ( (devParam = new DevTest::DevParam) == NULL )
			return 1;
		devArgList.push_back(devParam);
	}
#endif

	for ( it = devArgList.begin(); it != devArgList.end(); it++ )
	{
		DevTest::DevParam		*devParam;
		BOOL					setDevName;
		const char				*desc;

		devParam = *it;
		setDevName = FALSE;
		if ( devTest.srcDevParam == NULL )
			devTest.srcDevParam = devParam;
		else if ( devTest.dstDevParam == NULL )
			devTest.dstDevParam = devParam;

		desc = devParam->GetDesc();
#if	defined(WIN32)
		if ( desc == NULL )
		{	// Scan

		} else if ( (desc[1] == ':' || desc[1] == '\0') ) // drive letter
		{
			if ( isalpha(desc[0]) )
			{
				sprintf(devtmp, "\\\\.\\%c:", toupper(desc[0]));
				devParam->SetDevName(devtmp);
				setDevName = TRUE;
			} else
			{
				devTest.opt.cmd      = DevTest::CMD_SCAN;
				devTest.opt.scanMode = DevTest::SCAN_DRIVE;
			}
		} else if ( _strnicmp(desc, "disk", 4) == 0 )	// disk
		{
			if ( isdigit(desc[4]) )
			{
				sprintf(devtmp, "\\\\.\\PhysicalDrive%d", (int)atoi(&desc[4]));
				devParam->SetDevName(devtmp);
				setDevName = TRUE;
			} else
			{
				devTest.opt.cmd      = DevTest::CMD_SCAN;
				devTest.opt.scanMode = DevTest::SCAN_DISK;
			}
		} else if ( _strnicmp(desc, "scsi", 4) == 0 )	// scsi
		{
			if ( isdigit(desc[4]) )
			{
				sprintf(devtmp, "\\\\.\\scsi%d:", (int)atoi(&desc[4]));
				devParam->SetDevName(devtmp);
				setDevName = TRUE;
			} else
			{
				devTest.opt.cmd      = DevTest::CMD_SCAN;
				devTest.opt.scanMode = DevTest::SCAN_SCSI;
			}
#if	defined(DEVTEST_USE_CDROM)
		} else if ( _strnicmp(desc, "cdrom", 5) == 0 )	// CD-ROM
		{
			if ( isdigit(desc[5]) )
			{
				sprintf(devtmp, "\\\\.\\CdRom%d", atoi(&desc[5]) );
				devParam->SetDevName(devtmp);
				setDevName = TRUE;
			} else
			{
				devTest.opt.cmd      = DevTest::CMD_SCAN;
				devTest.opt.scanMode = DevTest::SCAN_CDROM;
			}
#endif
#if	defined(DEVTEST_USE_MD4RNET)
		} else if ( _strnicmp(desc, "net", 3) == 0 )	// MD4RAM NETWORK DEVICE
		{	// net:<host>:<port>:<pathid>:<targetid>:<lun>
			const char	*s;
			char		host[256];
			int			port, pathId, targetId, lun;

			s = desc;
			if ( DevTest::GetScsiPath(&s, sizeof(host)/sizeof(*host), host, &port, &pathId, &targetId, &lun) == 0 )
			{
				devParam->SetNet(host, port, pathId, targetId, lun);
				sprintf(devtmp, "\\\\.\\MD4RNet\\%s\\%d\\%d\\%d\\%d", host,  port, pathId, targetId, lun);
				devParam->SetDevName(devtmp);
				setDevName = TRUE;
			}
#endif
		} else
		{
			if ( strlen(desc) + 8 > sizeof(devtmp)/sizeof(*devtmp) )
				return -1;
			sprintf(devtmp, "\\\\.\\%s", desc );
			devParam->SetDevName(devtmp);
			setDevName = TRUE;
		}
		if ( setDevName != FALSE )
		{
			if ( devTest.opt.cmd == DevTest::CMD_SCAN )
				devTest.opt.cmd = DevTest::CMD_SHOW;
		} else
			devTest.opt.cmd = DevTest::CMD_SCAN;
#endif // end if ifdef WIN32

#if	defined(LINUX)
		if ( desc == NULL )
			continue;
		devParam->SetDevName(desc);
#endif

		switch ( devTest.opt.cmd )
		{
			case DevTest::CMD_SCAN:
#if	defined(WIN32)
				devTest.ScanDevice(*devParam);
#endif
				break;

			case DevTest::CMD_SHOW      :
			case DevTest::CMD_DUMP      :
			case DevTest::CMD_BENCHMARK :
			case DevTest::CMD_SURFSCAN  :
			case DevTest::CMD_READBLOCK :
#if	defined(DEVTEST_USE_DISCSPEED)
			case DevTest::CMD_DISCSPEED:
#endif
			case DevTest::CMD_MEDIUM:
				if ( devParam->GetDevName() == NULL )
					return 1;
				devTest.TargetDevice(*devParam);
				break;

		}
	} // end of for

	return 0;
}
