//////////////////////////////////////////////////////////////////////////
//	"varbuf.cpp" for md4rMan
//------------------------------------------------------------------------
//	$Id: varbuf.cpp 105 2006-09-20 21:36:16Z taro $
//////////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#	include "config.h"
#endif

#include	"varbuf.h"

#include	<stdarg.h>
#include	<stdio.h>

#ifndef	CRLF
# define	CRLF	TEXT("\r\n")
#endif

//////////////////////////////////////////////////////////////////////////
//	VarBuffer

VarBuffer::VarBuffer()
{
	this->buffersize = 0;
	this->validsize = 0;
	this->buffer = NULL;
}

VarBuffer::~VarBuffer()
{
	this->Clear();
}

void	VarBuffer::Clear(void)
{
	if ( this->buffer != NULL )
	{
		delete [] (BYTE *)this->buffer;
		this->buffer = NULL;
	}
	this->buffersize = 0;
	this->validsize = 0;
}

int		VarBuffer::Add(unsigned size, const void * data)
{
	if ( this->validsize + size >= this->buffersize )
	{
		unsigned	newsize;
		void	   *newbuffer;

		newsize = ((this->validsize + size + LINE_BUFFER_SIZE - 1) / LINE_BUFFER_SIZE) * LINE_BUFFER_SIZE;
		if ( (newbuffer = new BYTE [newsize]) == NULL )
			return RES_ERROR;
		if ( this->buffer != NULL )
		{
			if ( this->validsize > 0 )
				::memcpy( newbuffer, this->buffer, this->validsize);
			delete [] (BYTE *)this->buffer;
		}
		this->buffer     = newbuffer;
		this->buffersize = newsize;
	}
	if ( size > 0 && data != NULL )
	{
		::memcpy( (BYTE *)this->buffer + this->validsize, data, size);
		this->validsize += size;
	}

	return 0;
}

int		VarBuffer::AddChar(int ch)
{
	BYTE	b;

	if ( ch >= 256 )
	{
		b = (BYTE)(ch >> 8);
		if ( this->Add(sizeof(b), &b) < 0 )
			return RES_ERROR;
	}
	b = (BYTE)ch;
	return this->Add( sizeof(b), &b);
}

int		VarBuffer::Printf(const TCHAR * form, ... )
{
	va_list		arg;
	TCHAR		msg[4096];

	va_start(arg, form);
	::_vstprintf(msg, form, arg);
	va_end(arg);
	return this->AddText(msg);
}

void	VarBuffer::PutPos(int pos, ... )
{
	va_list	arg;
	TCHAR *	text;
	int		x = 0, len;

	va_start(arg, pos);
	do
	{
		text = va_arg(arg, TCHAR *);
		if ( text != NULL )
			len = (int)_tcslen(text);
		else
			len = 0;
		if ( pos > x )
		{
			this->Printf(TEXT("%*s"), pos - x, TEXT(""));
			x = pos;
		}
		this->Printf(TEXT("%s"), text);
		x += len;
		pos  = va_arg(arg, int);
	} while ( pos > 0 );
	if ( pos == 0 )
		this->Printf(CRLF);
	va_end(arg);
}

//////////////////////////////////////////////////////////////////////////
//	ǂݍ

int		VarBuffer::SetBuffer(void)
{
	this->p   = this->GetText();
	this->len = this->GetLength();
	this->pos = 0;
	this->ln = this->cn = 0;
	return 0;
}

int		VarBuffer::SetBuffer(unsigned len, const TCHAR * text)
{
	this->Clear();
	this->Add(len, text);
	return this->SetBuffer();
}

int		VarBuffer::IsEnd(void)
{
	return this->pos < this->len ? FALSE : TRUE;
}

int		VarBuffer::GetChar(void)
{
	int			ch;

	if ( this->IsEnd() )
		return RES_EOF;
	ch = this->p[this->pos++] & 0xFF;
	if ( _ismbblead(ch) && this->IsEnd() == FALSE )
	{
		int		ch2;

		ch2 = this->p[this->pos++] & 0xFF;
		if ( _ismbbtrail(ch2) )
			ch = (ch << 8) | ch2;
		else
			this->pos--;
	}

	if ( ch == '\n' )
	{
		this->ln++;
		this->cn = 0;
	}
	return ch;
}

int		VarBuffer::UnGetChar(int ch)
{
	if ( ch < 0 )
		return -1;
	if ( ch == '\n' )
		this->ln--;
	this->pos--;
	if ( ch >= 256 )
		this->pos--;
	return 0;
}

int		VarBuffer::SkipSpace(void)
{
	int			ch;

	for (;;)
	{
		if ( (ch = this->GetChar()) < 0 )
			return 0;
		switch ( ch )
		{
			case TEXT(' '):
			case TEXT('\t'):
			case TEXT('\r'):
			case TEXT('\n'):
				break;
#if 0
			case TEXT('/'):
			{
				int		ch2;

				if ( (ch2 = this->GetChar()) < 0 )
				{
					this->UnGetChar(ch);
					return 0;
				}
				if ( ch == TEXT('/') || ch == TEXT('*') )
				{
					this->SkipComment(ch);
				} else
				{
					this->UnGetChar(ch2);
					return 0;
				}
				break;
			}
#endif
			default:
				this->UnGetChar(ch);
				return 0;
		}
	}
}

int		VarBuffer::GetWord(VarBuffer & word)
{
	int			ch;
	unsigned	size = 0;

	this->SkipSpace();
	if ( (ch = this->GetChar()) < 0 )
		return ch;
	for(;;)
	{
		if ( word.AddChar(ch) < 0 )
			return RES_ERROR;
		size++;
		if ( (ch = this->GetChar()) < 0 )
			break;
		if ( this->IsNameChar(ch) == FALSE )
		{
			this->UnGetChar(ch);
			break;
		}
	}
	if ( word.AddChar('\0') < 0 )
		return RES_ERROR;

	return size;
}

int		VarBuffer::IsSpace(int ch)	// static/const
{
	return (ch == TEXT(' ') || ch == TEXT('\n') || ch == TEXT('\r') || ch == TEXT('\t')) ? TRUE : FALSE;
}

int		VarBuffer::IsNameTopChar(int ch)	// static/const
{
	if ( (ch >= TEXT('A') && ch <= TEXT('Z'))
	  || (ch >= TEXT('a') && ch <= TEXT('z'))
	  || ch == TEXT('_') )
		return 1;
	return 0;
}

int		VarBuffer::IsNameChar(int ch) // static/const
{
	if ( (ch >= TEXT('A') && ch <= TEXT('Z'))
	  || (ch >= TEXT('a') && ch <= TEXT('z'))
	  || (ch >= TEXT('0') && ch <= TEXT('9'))
	  || ch == TEXT('_') )
		return 1;
	return 0;
}

int		VarBuffer::GetString(VarBuffer & str, int term)
{
	int			ch;

	str.Clear();
	for(;;)
	{
		if ( (ch = this->GetChar()) < 0 )
			return ch;
		if ( ch == term )
			break;
		if ( str.AddChar(ch) < 0 )
			return RES_ERROR;
	}
	if ( str.AddChar('\0') < 0 )
		return RES_ERROR;
	return str.GetSize();
}

int		VarBuffer::GetLine(VarBuffer & str)
{
	int			ch;

	str.Clear();
	for(;;)
	{
		if ( (ch = this->GetChar()) < 0 )
			return ch;
		if ( ch == TEXT('\r') )
			continue;
		if ( ch == 0 || ch == TEXT('\n') )
			break;
		if ( str.AddChar(ch) < 0 )
			return RES_ERROR;
	}
	if ( str.AddChar('\0') < 0 )
		return RES_ERROR;
	return str.GetSize();
}


#if	defined(VARBUF_DEBUG)
#include	<stdio.h>
int	main(int argc, char **argv)
{
	VarBuffer	vbuf;

	{
		static char	data[] = "abcdefg Test String Taro Jiro";
		int					i;

		for ( i = 0; i < sizeof(data); i++ )
			vbuf.AddChar(data[i]);
		vbuf.AddChar('\0');
	}
	{
		fprintf(stdout, "Buffer Size %u \n", vbuf.buffersize);
		fprintf(stdout, "Valid Size %u \n", vbuf.validsize);
		fprintf(stdout, "Text % s\n", vbuf.GetText());
	}

#if 0
	{
		FILE				*fp;
		const char *		filename = "..\\icsvarbuf.cpp";

		if ( (fp = fopen(filename,"r")) == NULL )
			return 1;	// error
		fclose(f):
	}
#endif

	return 0;
}
#endif
