//////////////////////////////////////////////////////////////////////////
//	"dev_openclose.cpp"
//------------------------------------------------------------------------
//	$Id: dev_openclose.cpp 94 2006-09-20 08:51:21Z taro $
//////////////////////////////////////////////////////////////////////////

#if (_WIN32_WINNT < 0x500)
# undef		_WIN32_WINNT
# define	_WIN32_WINNT	0x500
#endif

#include	<winsock2.h>
#include	"devctl.h"
#include	<tchar.h>
#include	<stdio.h>
#if	defined(DEVCTL_USE_MD4RNET)
# include	"../md4r_sock.h"
#endif

#if defined(_DEBUG)
# include	<crtdbg.h>
# define	DBGREPORT(param_)	_CrtDbgReport param_
# define	DBGPRINT(param_)	printf param_
#else
# define	DBGREPORT(param_)
# define	DBGPRINT(param_)
#endif

int		DevCtl::OpenDevice(const TCHAR *name, unsigned rwmode)
{
	DWORD	dwAcc, dwShare, dwAtt;

	DBGREPORT((_CRT_WARN,NULL,NULL,NULL, "DevCtl::OpenDevice(%s, %08X)\n", name, rwmode));
	if ( this->handle != NULL )
		return -1;
	if ( this->devname != NULL )
		delete [] this->devname;
	if ( (this->devname= new TCHAR [_tcslen(name)+1]) == NULL )
		return -1;
	_tcscpy(this->devname, name);

	this->htype = HT_UNKNOWN;
	if ( _tcsnicmp(this->devname, TEXT("\\\\.\\PhysicalDrive"), 17) == 0 )
		this->htype = HT_DISK;
	else if ( _tcsnicmp(this->devname, TEXT("\\\\.\\"), 3) == 0
	  && isalpha(this->devname[4]) && this->devname[5] == ':' && this->devname[6] == '\0' )
		this->htype = HT_DRIVE;
	else if ( _tcsnicmp(this->devname, TEXT("\\\\.\\scsi"), 8) == 0 )
		this->htype = HT_SCSI;
#if	defined(DEVCTL_USE_MD4RNET)
	else if ( _tcsnicmp(this->devname, TEXT("\\\\.\\MD4RNet"), 11) == 0 )
	{	//	MD4RNet\\<host>:<port>-<pathId>-<targetId>-<lun>
		if ( this->socklib == NULL
		  || this->socklib->devwork.hostname == NULL
		  || *(this->socklib->devwork.hostname) == '\0'
		  || this->socklib->devwork.port < 0 )
		{
			TCHAR		host[256];
			const TCHAR	*s = &(this->devname[11]);
			int			port;

			if ( *s == TEXT('\\') )
				++s;
			if ( MD4R_Sock::GetNetDrivePath(&s, sizeof(host)/sizeof(*host), host, &port, NULL, NULL, NULL) < 0 )
				return -1;
			if ( this->NetSet(host, port) < 0 )
				return -1;
		}
		this->htype = HT_NET;
	}
#endif

	dwAcc = dwShare = dwAtt = 0;
	if ( rwmode & RW_READ )
		dwAcc |= GENERIC_READ;
	if ( rwmode & RW_WRITE )
		dwAcc |= GENERIC_WRITE;
	if ( rwmode & RW_SHARE_READ )
		dwShare |= FILE_SHARE_READ;
	if ( rwmode & RW_SHARE_WRITE )
		dwShare |= FILE_SHARE_WRITE;

	if ( rwmode & RW_ASYNC )
	{
		DBGREPORT((_CRT_WARN,NULL,NULL,NULL, "    Async Mode\n"));
		dwAtt |= FILE_FLAG_OVERLAPPED;
	}
	if ( rwmode & RW_NOBUFFER )
	{
		DBGREPORT((_CRT_WARN,NULL,NULL,NULL, "    No Buffering Mode\n"));
		dwAtt |= FILE_FLAG_NO_BUFFERING;
	}

	this->rwmode = rwmode;

#if	defined(DEVCTL_USE_MD4RNET)
	if ( this->htype == HT_NET )
		return 0;
#endif

	if ( (this->handle = ::CreateFile(devname, dwAcc, dwShare, NULL, OPEN_EXISTING, dwAtt, NULL)) == INVALID_HANDLE_VALUE )
	{
		this->handle = NULL;
		return -1;
	}
	DBGREPORT((_CRT_WARN,NULL,NULL,NULL, "    handle = 0x%08X\n", this->handle));
	if ( (this->rwmode & RW_ASYNC) && this->hEvent == NULL )
	{
		this->hEvent = ::CreateEvent(NULL, TRUE, FALSE, NULL);
		DBGREPORT((_CRT_WARN,NULL,NULL,NULL, "    hEvent = 0x%08Xe\n", this->hEvent));
	}

	return 0;
}

int		DevCtl::CloseDevice(void)
{
	DBGREPORT((_CRT_WARN,NULL,NULL,NULL, "DevCtl::CloseDevice(): handle = 0x%08X\n", this->handle));
#if defined(DEVCTL_USE_FILEIO)
	if ( this->ctl != NULL )
	{
		delete this->ctl;
		this->ctl = NULL;
	}
#endif
#if	defined(DEVCTL_USE_MD4RNET)
	if ( this->htype == HT_NET )
	{
		this->DestroySock();
		this->htype = HT_UNKNOWN;
		return 0;
	}
#endif
	if ( this->handle == NULL )
		return -1;
	::CloseHandle(this->handle);
	this->handle = NULL;

	if ( this->hEvent != NULL )
	{
		::CloseHandle(this->hEvent);
		this->hEvent = NULL;
	}

	return 0;
}

