//////////////////////////////////////////////////////////////////////////
//	"net_openclose.cpp"
//------------------------------------------------------------------------
//	$Id: net_openclose.cpp 107 2006-09-20 23:51:27Z taro $
//////////////////////////////////////////////////////////////////////////

#if (_WIN32_WINNT < 0x500)
# undef		_WIN32_WINNT
# define	_WIN32_WINNT	0x500
#endif

#include	<winsock2.h>
#include	"../md4r_sock.h"
#include	"../crc16.h"
#include	"../sb8crc.h"

#include	"devctl.h"
#include	<tchar.h>
#include	<stdio.h>
#include	<stddef.h>
# include	<crtdbg.h>

#if defined(_DEBUG)
# define	DBGREPORT(param_)	_CrtDbgReport param_
# define	DBGPRINT(param_)	printf param_
#else
# define	DBGREPORT(param_)
# define	DBGPRINT(param_)
#endif

#if	defined(DEVCTL_USE_MD4RNET) && defined(WIN32)
void	DevCtl::DestroySock(void)
{
	if ( this->socklib != NULL )
	{
		delete this->socklib;
		this->socklib = NULL;
	}
	return;
}

int			DevCtl::NetSet(const TCHAR * host, int port)
{
	if ( this->handle != NULL )
		return -1;
	if ( this->socklib == NULL )
	{
		if ( (this->socklib = new MD4R_Sock) == NULL )
			return -1;
	}
	this->socklib->SetHostName(host);
	this->socklib->devwork.port = port;
	this->socklib->devwork.hostaddr.SetAddr((const char *)host);
	this->socklib->devwork.hostaddr.SetPort(port);
	this->handle = NULL;
	this->htype  = HT_NET;

	return 0;
}

int			DevCtl::NetOpen(unsigned rwmode)
{
	if ( this->socklib == NULL )
		return -1;
	if ( this->socklib->Create(SOCK_STREAM) < 0 )
		return -1;
	if ( this->socklib->Connect(this->socklib->devwork.hostaddr) < 0 )
		return -1;

	{
		int		tm;

		if ( this->socklib->recv_timeout > 0 )
		{
			tm = this->socklib->recv_timeout;
			this->socklib->SetSockOpt(SOL_SOCKET, SO_RCVTIMEO, &tm, sizeof(tm));

		}
		if ( this->socklib->send_timeout > 0 )
		{
			tm = this->socklib->send_timeout;
			this->socklib->SetSockOpt(SOL_SOCKET, SO_SNDTIMEO, &tm, sizeof(tm));
		}
	}
	return 0;
}

int			DevCtl::NetClose(void)
{
	if ( this->socklib == NULL )
		return -1;
	if ( this->socklib->IsCreate() )
		this->socklib->Close();

	return 0;
}

//	scdb->dataBufferOffset͊֐ŃZbg
int			DevCtl::NetScsiSend(MD4R_NetScdb * scdb, unsigned outSize, void * outData)
{
	MD4R_NetHead	head;
	unsigned		headLength, head_dataLength, head_dataOffset;
	unsigned		senseLength, sendOffset = 0;

	_ASSERT(scdb != NULL);
	memset(&head, 0, sizeof(head));
	memcpy(head.chead.id     , MD4R_Sock::md4r_net_id     , sizeof(head.chead.id     ));
	memcpy(head.chead.version, MD4R_Sock::md4r_version_id , sizeof(head.chead.version));
	memcpy(head.chead.marker , MD4R_Sock::md4r_marker_scdb, sizeof(head.chead.marker ));
	headLength = sizeof(head);
	SCSI_PUT32(head.chead.length, headLength);

	head.att.headCrc = 1;
	head.att.dataCrc = 1;

	senseLength = scdb->senseInfoLength;
	if ( senseLength < MD4R_NET_SCSI_SENSE_LENGTH )
		senseLength = MD4R_NET_SCSI_SENSE_LENGTH;

	if ( outSize > 0 )
	{
		sendOffset = sizeof(MD4R_NetScdb) + senseLength;
		SCSI_PUT32(scdb->dataBufferOffset  , sendOffset );
	}

	head_dataLength = sizeof(MD4R_NetScdb) + senseLength + outSize;
	SCSI_PUT32(head.dataLength  , head_dataLength);
	head_dataOffset = sizeof(head);
	SCSI_PUT32(head.dataOffset  , head_dataOffset);

	if ( head.att.headCrc )
	{
		Crc16work	crcwork;
		crc16_t		crc;

		crc16xm_init(&crcwork);
		crc16xm_calc(&crcwork, offsetof(MD4R_NetHead, headCrc16), &head);
		crc = crc16xm_getcrc(&crcwork);
		crc16xm_term(&crcwork);
		SCSI_PUT16(head.headCrc16, crc);
	}

	// wb_M
	if ( this->socklib->Write(sizeof(head), &head) < sizeof(head) )
		return -1;
	// SCSIR}hM
	headLength = sizeof(MD4R_NetScdb) + senseLength;
	if ( this->socklib->Write(headLength, scdb) < (int)headLength )
		return -1;
	// f[^o
	if ( outSize > 0 && outData != NULL )
	{
		if ( this->socklib->Write(outSize, outData) < (int)outSize )
			return -1;
	}
	// END OF MARKo
	{
		MD4R_NetEom	eom;

		memset(&eom, 0, sizeof(&eom));
		memcpy(eom.chead.id     , MD4R_Sock::md4r_net_id     , sizeof(eom.chead.id     ));
		memcpy(eom.chead.version, MD4R_Sock::md4r_version_id , sizeof(eom.chead.version));
		memcpy(eom.chead.marker , MD4R_Sock::md4r_marker_end , sizeof(eom.chead.marker ));
		SCSI_PUT32(eom.chead.length, sizeof(eom));
		if ( head.att.dataCrc )
		{
			Sb8crcwork	crcwork;
			sb8crc_t	crc;

			eom.att.dataCrc = 1;
			sb8crc_init(&crcwork);
			sb8crc_calc(&crcwork, headLength, scdb);
			if ( outSize > 0 && outData != NULL )
				sb8crc_calc(&crcwork, outSize, outData);
			crc = sb8crc_getcrc(&crcwork);
			sb8crc_term(&crcwork);
			SCSI_PUT32(eom.dataCrc32, crc);
		} else
			eom.att.dataCrc = 0;
		if ( this->socklib->Write(sizeof(eom), &eom) < sizeof(eom) )
			return -1;
	}

	return 0;
}

int			DevCtl::NetScsiRecv(MD4R_NetScdb * scdb, unsigned senseSize, void * senseData, unsigned inpSize, void * inpBuffer)
{
	MD4R_NetHead	head;
	unsigned		pos = 0;
	unsigned		headLength, subLength, dataLength;
	int				subOffset, dataOffset;
	int				ret, retcode = 0;
	BYTE			*data = NULL;

	try
	{
		sb8crc_t		crc;

		ret = this->socklib->ReadHeader(&head.chead, sizeof(head));
		if ( ret < sizeof(head) )
			throw -1;
		pos += ret;
	//	headLength = MD4R_GET32(head.chead.length);
		if ( head.att.headCrc )
		{	// CRC
			Crc16work	crcwork;
			crc16_t		crc, headCrc;

			crc16xm_init(&crcwork);
			crc16xm_calc(&crcwork, offsetof(MD4R_NetHead, headCrc16), &head);
			crc = crc16xm_getcrc(&crcwork);
			crc16xm_term(&crcwork);
			headCrc = SCSI_GET16(head.headCrc16);
			if ( headCrc != crc )
			{
				DBGPRINT(("DEBUG: Header crc error!\n"));
			}
		}
		subLength    = MD4R_GET32(head.subLength);
		subOffset    = MD4R_GET32(head.subOffset);
		dataLength   = MD4R_GET32(head.dataLength);
		dataOffset   = MD4R_GET32(head.dataOffset);
		if ( subLength > 0 && subOffset >= (int)pos )
		{	// sub header
			unsigned	dummy_size;

			if ( (dummy_size = subOffset - pos) > 0 )
			{
				if ( (ret = this->socklib->ReadDummy(dummy_size)) < (int)dummy_size )
					throw -1;
				ret += pos;
			}
			if ( pos != subOffset )
				throw -1;
			// read sub header
			if ( (ret = this->socklib->ReadDummy(subLength)) < (int)subLength )
				throw -1;
			ret += pos;
		}
		if ( dataLength > 0 )
		{	// Data body
			unsigned	dummy_size;

			if ( (dummy_size = dataOffset - pos) > 0 )
			{
				if ( (ret = this->socklib->ReadDummy(dummy_size)) < (int)dummy_size )
					throw -1;
				ret += pos;
			}
			if ( pos != dataOffset && dataLength > MD4R_NET_MAX_DATA_LENGTH )
				throw -1;
			if ( (data = new BYTE [dataLength]) == NULL )
				throw -1;
			if ( (ret = this->socklib->Read(dataLength, data)) < (int)dataLength )
				throw -1;
			ret += pos;
			if ( head.att.dataCrc )
			{
				Sb8crcwork	crcwork;

				sb8crc_init(&crcwork);
				sb8crc_calc(&crcwork, dataLength, data);
				crc = sb8crc_getcrc(&crcwork);
				sb8crc_term(&crcwork);
			}
		}

		// END OF MARK 
		{
			MD4R_NetEom		eom;
			unsigned		eomLength;

			ret = this->socklib->ReadHeader(&eom.chead, sizeof(eom));
			if ( ret < sizeof(eom) )
				throw -1;
		//	eomLength = MD4R_GET32(eom.chead.length);
			if ( head.att.dataCrc && eom.att.dataCrc )
			{
				sb8crc_t		crcData;

				crcData = SCSI_GET32(eom.dataCrc32);
				if ( crc != crcData )
				{
					DBGPRINT(("DEBUG: data crc error!\n"));
				}
			}
		}

		if ( memcmp(head.chead.marker, MD4R_Sock::md4r_marker_scdb, MD4R_NET_MARKER_LENGTH) == 0  )
		{	// SCSI Command
			if ( dataLength >= sizeof(MD4R_NetScdb) && data != NULL )
			{
				MD4R_NetScdb	*scdb;
				unsigned		senseLength;

				scdb = (MD4R_NetScdb *) data;
				if ( SCSI_GET16(scdb->length) < sizeof(*scdb) )
					throw -1;
				switch ( scdb->scsiStatus )
				{
				case DevCtl::SCSI_STATUS_GOOD:
					break;	// OK
				case DevCtl::SCSI_STATUS_CHECK_CONDITION:
					if ( (senseLength = scdb->senseInfoLength) > 0 && senseSize > 0 && senseData != NULL )
					{
						BYTE	*src;
						int		offset;

						if ( senseLength > senseSize )
							senseLength = senseSize;
						offset = SCSI_GET32(scdb->senseInfoOffset);
						if ( offset < 0 || offset + senseLength > dataLength )
							throw -1;
						src = ((BYTE *)scdb) + offset;
						memcpy(senseData, src, senseLength);
					}
				default:
					throw 0;
				}
				switch ( scdb->dataIn )
				{
				case MD4R_NET_DATA_IN :	// T[óif[^Mj
					if ( data != NULL && dataLength > 0 )
					{
						BYTE		*dataBuffer;
						unsigned	dataTransferLength;
						int			offset;

						dataTransferLength = SCSI_GET32(scdb->dataTransferLength);
						if ( dataTransferLength > inpSize )
							dataTransferLength = inpSize;
						offset = SCSI_GET32(scdb->dataBufferOffset);
						if ( offset < 0 || offset + dataTransferLength > dataLength )
							throw -1;
						dataBuffer = ((BYTE *)scdb) + offset;
						memcpy(inpBuffer, dataBuffer, dataTransferLength);
					}
					break;

				case MD4R_NET_DATA_OUT:	// T[oɏo
				//	outspt->DataTransferLength = SCSI_GET32(scdb->dataTransferLength);
					break;
				}
			}
			retcode = 0;
		} else
		{
			DBGPRINT(("DEBUG: marker mismatch!\n"));
			retcode = -1;
		}
	} catch(int ret) { retcode = ret; }

//	this->socklib->Close();
	if ( data != NULL )
		delete [] data;

	return retcode;
}

#endif
