/////////////////////////////////////////////////////////////////////////
//	"udflib.h"
//------------------------------------------------------------------------
//	$Id: udflib.h 181 2006-09-20 10:50:31Z taro $
//////////////////////////////////////////////////////////////////////////
#ifndef	UDFLIB_H_INCLUDED
#define	UDFLIB_H_INCLUDED

#define	UDFLIB_CFG_LE

#if	defined(WIN32)
# include	<win32/devctl.h>
#endif
#if	defined(LINUX)
# include	<linux/devctl.h>
#endif

#if	!defined(ATTB_PACKED)
# if	defined(LINUX)
#   define	ATTB_PACKED	__attribute ((packed))
# endif
# if	defined(WIN32)
#   define	ATTB_PACKED
# endif
#endif

class	UdfLib
{
public:
	enum udflib_enum
	{
		MAX_VOL_EXTS = 100,
	};

#if	defined(_MSC_VER)
#	pragma pack(push,1)
#endif
	typedef unsigned char		Uint8;
	typedef signed char			Sint8;
	typedef unsigned short		Uint16;
	typedef signed short		Sint16;
	typedef unsigned int		Uint32;
	typedef int					Sint32;
#if	defined(LINUX)
	typedef unsigned long long	Uint64;
	typedef long long			Sint64;
#else
	typedef unsigned __int64	Uint64;
	typedef __int64				Sint64;
#endif
	typedef Uint8				dstring;
	typedef unsigned short		unicode_t;

	enum tagid_enum	// Uint16
	{
		TAGID_NONE             =  0   , // no tag
		TAGID_PVD              =  1   , // primary volume desc
		TAGID_ANCHOR           =  2   , // anchor desc
		TAGID_POINTER          =  3   , // pointer desc
		TAGID_IUVD             =  4   , // implementation use volume desc
		TAGID_PD               =  5   , // volume partition desc
		TAGID_LVD              =  6   , // logical volume desc
		TAGID_USD              =  7   , // unallocated volume space desc
		TAGID_TERM_DESC        =  8   , // terminator desc
		TAGID_LVID             =  9   , // logical volume integrity desc
		TAGID_FSD              =  256 , // file set desc
		TAGID_FILE_ID          =  257 , // file identifier desc
		TAGID_ALLOC_EXTENT     =  258 , // Allocation extent desc
		TAGID_INDIRECT         =  259 , // Indirect entry
		TAGID_TERM_ENTRY       =  260 , // Terminal entry
		TAGID_FILE_ENTRY       =  261 , // File entry
		TAGID_EXT_ATTR         =  262 , // Extended attribute desc
		TAGID_UNALLOC_SP_ENTRY =  263 , // Unallocated space entry (WORM)
		TAGID_SPACE_BMAP       =  264 , // Space bitmap desc
		TAGID_PART_INTEGRITY   =  265 , // Partition integrity desc
	};
	struct charspec	// ECMA 167 1/7.2.1 (64 bytes)
	{
		Uint8		charSetType;
		Uint8		charSetInfo[63];
	} ATTB_PACKED;
	struct extent_ad
	{
		Uint32		length;
		Uint32		location;
	} ATTB_PACKED;
	struct timestamp	// ECMA 167 1/7.3
	{
		Sint16		typeAndTimeZone;
		Sint16		year;
		Uint8		month;
		Uint8		day;
		Uint8		hour;
		Uint8		minute;
		Uint8		second;
		Uint8		centiseconds;
		Uint8		hundredMicroseconds;
		Uint8		microseconds;
	} ATTB_PACKED;
	struct regid	// Entity Indentifier ECMA 167 1/7.4 
	{
		Uint8		flags;
		Uint8		id[23];
		Uint8		idSuffix[8];
	} ATTB_PACKED;
	struct udfEntityId	// UDF1.01 / 2.1.4.2
	{
		Uint8		flags;
		Uint8		id[23];
		Uint16		UDFRevision;
		Uint8		OSClass;
		Uint8		OSIdentifier;
		Uint8		reserved[4];
	} ATTB_PACKED;
	struct implEntityId
	{
		Uint8		flags       ;
		Uint8		id[23]      ;
		Uint8		OSClass     ;
		Uint8		OSIdentifier;
		Uint8		implUse[6]  ;
	} ATTB_PACKED;
	struct domainEntityId	// 32 bytes
	{
		Uint8		flags;
		Uint8		id[23];
		Uint16		UDFRevision;
		Uint8		domainFlags;
		Uint8		rReserved[5];
	} ATTB_PACKED;
	struct LVInformation	// IVUD implementationUse
	{
		charspec		LVICharset;
		dstring			logVolId[128];
		dstring			LVInfo1[36];
		dstring			LVInfo2[36];
		dstring			LVInfo3[36];
		implEntityId	implementationID;
		Uint8			implementationUse[128];
	} ATTB_PACKED;
	struct VRS		// Volume Recognition Structures
	{
		Uint8		type;
		Uint8		sid[5];
		Uint8		version;
	} ATTB_PACKED;
	struct DTag	// DescripterTag (16 bytes)
	{
		Uint16		tagID;
		Uint16		descriptorVersion;
		Uint8		tagChecksum;
		Uint8		reserved;
		Uint16		tagSerialNum;
		Uint16		descriptorCrc;
		Uint16		crcLen;
		Uint32		tagLoc;
	} ATTB_PACKED;
	struct AVDP			// Anchor Volume Descriptor Pointer
	{
		DTag			tag;
		extent_ad		mainVDSAdr;
		extent_ad		reserveVDSAdr;
		Uint8			reserved[480];
	} ATTB_PACKED;
	struct PrimaryVolDes	// ECMA 167 3/10.1
	{
		DTag			tag;		// tagID = 1
		Uint32			volDescSeqNum;
		Uint32			primVolDesNum;
		dstring			volID[32];
		Uint16			volSeqNum;
		Uint16			maxVolSeqNum;
		Uint16			interchangeLevel;
		Uint16			maxInterchangeLevel;
		Uint32			charSetList;
		Uint32			maxCharSetList;
		dstring			volSetID[128];
		charspec		desCharSet;
		charspec		explanatoryCharSet;
		extent_ad		volAbstract;
		extent_ad		volCopyrightNotice;
		regid			applicationID;
		timestamp		recordingTime;
		implEntityId	implementationID;
		Uint8			implementationUse[64];
		Uint32			predecessorVDSLoc;
		Uint16			flags;
		Uint8			reserved[22];
	} ATTB_PACKED;
	struct LogVolDesc		// [3/10.6] Logical Volume Descriptor
	{
		DTag			tag;	// tagID = 6
		Uint32			volDescSeqNum;
		charspec		desCharSet;
		dstring			logVolID[128];
		Uint32			logBlkSize;
		domainEntityId	domainID;
		Uint8			logVolUse[16];
		Uint32			mapTabLen;
		Uint32			numPartMaps;
		implEntityId	implementationID;
		Uint8			implementationUse[128];
		extent_ad		integritySeqExtent;
	} ATTB_PACKED;
	struct VolDescPtr	// [3/10.3] Volume Descriptor Pointer
	{
		DTag			tag;	// tagID = 3
		Uint32			volDescSeqNum;
		extent_ad		nextVDS;
		Uint8			reserved[484];
	} ATTB_PACKED;
	struct ImpUseVolDesc	// ECMA [3/10.4] Implementation Use Volume Descriptor
	{
		DTag 			tag;	// tagID = 4
		Uint32			volDescSeqNum;
		udfEntityId		implementationID;
		Uint8			implementationUse[460];
	} ATTB_PACKED;
	struct UnallocSpDesHead		// ECMA 167 3/10.8 Unallocated Space Descriptor
	{
		DTag			tag;				// tagID = 7
		Uint32			volDescSeqNum;
		Uint32			numAllocationDes;
		extent_ad		allocationDes[1];
	} ATTB_PACKED;
	struct PartDesc	// ECMA 167 3/10.5 Partition Descriptor
	{
		DTag			tag;		// tagID = 5
		Uint32			volDescSeqNum;
		Uint16			partFlags;
		Uint16			partNumber;
		regid			partContents;
		Uint8			partContentsUse[128];
		Uint32			accessType;
		Uint32			partStartingLoc;
		Uint32			partLength;
		implEntityId	implementationID;
		Uint8			implementationUse[128];
		Uint8			reserved[156];
	} ATTB_PACKED;
	struct lb_addr		// ECMA 167 4/7.1 Recorded Address (6 bytes)
	{
		Uint32			logicalBlockNum;
		Uint16			partitionReferenceNum;
	} ATTB_PACKED;
	struct long_ad		// ECMA 167 4/14.14.2 Long Allocation Descriptor (16 bytes)
	{
		Uint32			extentLength;
		lb_addr			extentLocation;
		Uint8			implementationUse[6];
	} ATTB_PACKED;
	struct FileSetDescriptor	// ECMA 167 4/14.1 File Set Descriptor
	{
		DTag			tag;								// + 0x000
		timestamp		recordingDateAndTime;				// + 0x010
		Uint16			interchangeLevel;					// + 0x01C
		Uint16			maxInterchangeLevel;				// + 0x01E
		Uint32			charSetList;						// + 0x020
		Uint32			maxCharSetList;						// + 0x024
		Uint32			fileSetNumber;						// + 0x028
		Uint32			fileSetDescriptorNumber;			// + 0x02C
		charspec		logicalVolumeIdentifierChractorSet;	// + 0x030
		dstring			logicalVolumeIdentifier[128];		// + 0x070
		charspec		fileSetCharactorSet;				// + 0x0F0
		dstring			fileSetIdentifier[32];				// + 0x130
		dstring			copyrightFileIdentifier[32];		// + 0x150
		dstring			abstractFileIdentifier[32];			// + 0x170
		long_ad			rootDirectoryICB;					// + 0x190
		domainEntityId	domainIdentifier;					// + 0x1A0
		long_ad			nextExtent;							// + 0x1C0
		long_ad			streamDirectoryICB;					// + 0x1D0
		Uint8			reserved[32];						// + 0x1E0
	} ATTB_PACKED;
#if	defined(_MSC_VER)
#	pragma pack(pop)
#endif

	static Uint32	endian32(Uint32 toswap) { return (toswap << 24) | ((toswap << 8) & 0x00ff0000) | ((toswap >> 8) & 0x0000ff00) | (toswap >> 24); }
	static Uint16	endian16(Uint16 toswap) { return (toswap << 8) | (toswap >> 8); }
#if	defined(UDFLIB_CFG_LE)
	// For little endian machines (x86)
	static Sint32	S_endian32(Uint32 x) { return UdfLib::endian32(x); }
	static Sint16	S_endian16(Uint16 x) { return UdfLib::endian16(x); }
	static Uint32	U_endian32(Uint32 x) { return x; }
	static Uint16	U_endian16(Uint16 x) { return x; }
#else
	// For big endian machines
	static Sint32	S_endian32(Uint32 x) { return x; }
	static Sint16	S_endian16(Uint16 x) { return x; }
	static Uint32	U_endian32(Uint32 x) { return UdfLib::endian32(x); }
	static Uint16	U_endian16(Uint16 x) { return UdfLib::endian16(x); }
#endif

public:
	enum sid_enum
	{
		SID_BEA01,
		SID_BOOT2,
		SID_CD001,
		SID_CDW02,
		SID_NSR02,
		SID_NSR03,
		SID_TEA01,
		SID_NUM  ,
	};
	class VRSInfo
	{
	public:
		unsigned	size;
		struct Vrs
		{
			Vrs		   *next;
			unsigned	type;
			unsigned	loc;
			VRS			vrs;
		};
		unsigned		numVrs;
		Vrs				*vrsTop;
		Vrs				*vrsLast;
		VRSInfo()  { this->numVrs = 0; this->vrsTop = this->vrsLast = NULL; }
		~VRSInfo() { this->Clear(); }
		void		Clear();
		int			Add(unsigned type, unsigned loc, VRS * vrs);
	};
	struct ADVPInfo
	{
		unsigned		size;
		unsigned		flag;
		struct tagAdvp
		{
			INT64		loc;
			AVDP		avdp;
		} p[2];						// 0:front, 1:back
	};
	enum voldesc_enum
	{
		VOLDES_UNKNOWN,
		VOLDES_PVD ,
		VOLDES_LVD ,
		VOLDES_IUVD,
		VOLDES_USD ,
		VOLDES_PD  ,
	};
	struct	VolDes
	{
		VolDes				   *next;
		unsigned				type;
		unsigned				loc, len;
		union Des
		{
			DTag				tag;
			PrimaryVolDes		pvd;
			LogVolDesc			lvd;
			ImpUseVolDesc		iuvd;
			UnallocSpDesHead	usd;
			PartDesc			pd;
		};
		Des *					des;
	};
	class	VolDesInfo
	{
	friend class	UdfLib;
	public:
		unsigned	numDes;
		VolDes		*desTop;
		VolDes		*desLast;

		unsigned	volspcListLen;
		struct tagVolSpc
		{
			unsigned	location;
			unsigned	length;
			char		*name;
		} volspc[MAX_VOL_EXTS];

		VolDesInfo()
		{
			this->numDes = 0;
			this->desTop = this->desLast = NULL;
			this->volspcListLen = 0;
			memset(&this->volspc, 0, sizeof(this->volspc));
		}
		~VolDesInfo() { this->Clear(); }
		void		Clear(void);
		int			AddDes(unsigned type, unsigned loc, unsigned len, unsigned size, DTag * tag);
	};
	struct VdsInfo
	{
		VolDesInfo	mVolDes;
		VolDesInfo	rVolDes;
	};

public:
	UdfLib();
	~UdfLib();

	int		OpenUdf(DevCtl & devCtl, DevCtl::ScsiInqData & inq, DevCtl::SCSICAPACITYDATA16 & capa, INT64 sec_off, INT64 sec_num);
	int		CheckVRS(UdfLib::VRSInfo & vrsInfo);
	int		GetAVDP(UdfLib::ADVPInfo & avdpInfo);
	int		GetVDS(UdfLib::VdsInfo & vdsInfo, UdfLib::AVDP & avdp);
	int		ReadVDS(UdfLib::VolDesInfo & volDesInfo, Uint8 * buf, Uint32 location, Uint32 length);

private:
	DevCtl						*devCtl;
	DevCtl::ScsiInqData			*inqData;
	DevCtl::SCSICAPACITYDATA16	*capa;
	INT64						sec_off;
	INT64						sec_num;
	virtual int			ReadSector(INT64 adr, unsigned count, unsigned size, void * buffer);
public:
	virtual int			GetBlkSize(void);
	virtual INT64		GetBlkNum(void);

	int 	CheckTag(UdfLib::DTag & tag, Uint32 toaLoc, Uint16 tagId, int crc_min, int crc_max);
	Uint16	doCrc(UdfLib::Uint8 * buffer, int n);
	int 	track_volspace(UdfLib::VolDesInfo volDesInfo, Uint32 location, Uint32 length, char *name);
	int		CheckRegid(UdfLib::udfEntityId * reg, char * id);
	static int 	UncompressUnicode(int numberOfBytes, UdfLib::Uint8 *UDFCompressed, UdfLib::unicode_t *unicode);
};

#endif
